/****************************************************************************
 * This file is part of the program ALIFRITZ                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <string>
#include <vector>
#include <list>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <climits>  
#include <cfloat>  
#include <algorithm>  

using namespace std;

#include "random.hh"
#include "seqcode.hh"
#include "data.hh"
#include "substmod.hh"
#include "indel.hh"
#include "treespec.hh"
#include "profalign.hh"
#include "parser.hh"
#include "eingabe.hh"
#include "mult_align.hh"

#define FALSE 0
#define TRUE 1
#define FASTMODE
#define MINDESTENS 1
#define MAXMULT 10  
// eine gesampelte innere Sequenz darf hoechstens MAXMULT mal so lang sein, wie ihr laengster Nachbar
#define RSHIFT 10000.0/9999.0
// RSHIFT bestimmt das maximal zulaessige TKF2-r (10000.0/9999.0 entspr. r_max=0.9999)
#define DEBUG
#define INFACC 100.0

/*************************************************************/
/****************** class mult_alignment   *******************/
/*************************************************************/


/************** publike Funktionen             ***************/

/***************** Konstruktoren und Destruktor **************/

mult_alignment::mult_alignment()
{
  cerr << "You tried to construct an uninitialized object of the class mult_alignment.\n";
  exit(EXIT_FAILURE);
}


mult_alignment::mult_alignment(eingabe_parameter ep, int antyp)
{
   vector<double> frequs;
   alphsize alf;
   int format=ep.fil.format;
   
   // Sequenzen einlesen
   ifstream seqfile(ep.fil.infile.c_str());
   if(!seqfile) {
      cerr << "Cannot open sequence file "<<ep.fil.infile.c_str()<<endl;
      exit(EXIT_FAILURE);
   }
   dataset d(seqfile,format);
   if(d.spez()<2) {
      cerr << "There are " << d.spez() << " sequences.\n"
           << "There is nothing to align.\n";
      exit(EXIT_FAILURE);
   }  
   if(ep.sub.freqset)         // Userwunsch
      frequs = ep.sub.freqs;
   else
      d.getfreqs(frequs);
   seqfile.close();
   d.guesstype();
   alf=d.typ;
   

   // Variablen setzen
   sub = substmodel(alf,ep.sub.model,frequs,ep.sub.params);
   id  = indel(ep.ind.params,ep.ind.model);
   kee = ep.kee;
   coo = ep.coo;

   vector<vector<double> > distmat;
   if(antyp==1 || antyp==2) {
      cout<<"Computing pairwise distances in an approximate EM step\n";
      compute_distmat(d,distmat);
   }
   // Baum einlesen
   clustal bstr(ep.fil.guidetree.c_str());
   told = new baum(bstr,TRUE,d);
   // Baum mit einem Blatt wurzeln
   told=told->reroot();
   told->fuseblindnodes(TRUE); // verhindert *-*-*
   
   
   if(antyp==1 || antyp==2) {
      unsigned int taxcount=distmat.size();
      cout << "Building the initial NJ tree\n";
      told->NeighborJoining(distmat,taxcount);  // NJ-Baum unter Beruecksichtigung der User-Wuensche bzgl. der Topologie
   } else
     told->bifurcate(TRUE); // zufaelliger Baum

   tnew = NULL;
   zaehler = 0;
   // lnL wird am Schluss der Funktion gesetzt

   construct_align(told);
   uphill(told);// ein bisschen optimieren
   printform=mdataset(told,alf); // Alignment der Blaetter in ueblicher Form
   cosubi=compact_subst_pattern(printform); // Information ueber die Substitutionsmuster
   lnL = complik(cosubi,told);
optim_times();
   // Parameter fuers Simulated Annealing
   maxdifftree = 1.0;
   betatree = 0.0; // wird im letzten Burn-in Schritt gesetzt
   maxdiffparam = 1.0;
   betaparam = 0.0; // wird im letzten Burn-in Schritt gesetzt*/
   
   // Variablen zur Optimierung von MN-delta
   // Variablen zur Optimierung von MN-delta
   deltafaktor=8;
   MNstep=0;
   descending=TRUE;
   wieweit.resize(8);
   times_visited.resize(8);
   for(unsigned int i=0;i<8;++i) {
       wieweit[i]=0.0;
       times_visited[i]=0;
   }
}

mult_alignment::mult_alignment(const mult_alignment& mult1)
{
    sub=mult1.sub;
    id=mult1.id;
    kee=mult1.kee;
    coo=mult1.coo;
    told=new baum(*(mult1.told));
    tnew=NULL;
    printform=mult1.printform;
    cosubi=mult1.cosubi;
    lnL=mult1.lnL;

    zaehler=mult1.zaehler;
    maxdifftree=mult1.maxdifftree;
    betatree=mult1.betatree;
    maxdiffparam=mult1.maxdiffparam;
    betaparam=mult1.betaparam;

    // Variablen zur Optimierung von MN-delta
    deltafaktor=mult1.deltafaktor;
    MNstep=mult1.MNstep;
    descending=mult1.descending;
    wieweit=mult1.wieweit;
    times_visited=mult1.times_visited;
}


mult_alignment::~mult_alignment()
{
   if(told!=NULL) {
      if(!told->isRoot()) {
          told->reroot(told);
	  cerr << "WARNING: told's root has changed\n";
      }
      delete(told); // => Vorsicht beim Kopieren von mult_alignment
   }
   if(tnew!=NULL) {
      if(!tnew->isRoot()) {
          tnew->reroot(tnew);
	  cerr << "WARNING: tnew's root has changed\n";
      }
      delete(tnew);
   }
}


void mult_alignment::compute_distmat(dataset& d, vector<vector<double> >& distmat)
{
   unsigned int i, j, spez=d.spez();
   double anfangszeit=10.0;
   distmat.resize(spez);
   vector<vector<double> > scalemat(spez);
   for(i=0;i<spez;++i) {
       bioseq b1(d.Seqs[i],d.typ);
       distmat[i].resize(spez);
       scalemat[i].resize(spez);
       for(j=0;j<spez;++j) {
           if(j==i) {
	      distmat[i][j]=0.0;
	      scalemat[i][j]=1.0;
	   } else {
	      if(j<i) {
	         distmat[i][j]=distmat[j][i];
	         scalemat[i][j]=scalemat[j][i];
	      } else {
	         distmat[i][j]=anfangszeit;
	         scalemat[i][j]=1.0;
	         bioseq b2(d.Seqs[j],d.typ);
	         alignment pali(b1,b2,distmat[i][j],sub,id);
		 pali.optimal_time(distmat[i][j], scalemat[i][j]);
		 cout<<d.Identif[i]<<" "<<d.Identif[j]<<" distance="<<distmat[i][j]<<" scale="<<scalemat[i][j]<<endl;
	      }
	   }
       }
   }
   // lambda und mue anpassen
   id.rescale(median_scale(scalemat));
   cout << "Second round of EM algorithm.\nInsertion rate multiplied with "<<median_scale(scalemat)<<endl;

   // zweite Stufe der Optimierung
   for(i=0;i<spez;++i) {
       bioseq b1(d.Seqs[i],d.typ);
       for(j=0;j<spez;++j) {
           if(j==i) {
	      distmat[i][j]=0.0;
	   } else {
	      if(j<i) {
	         distmat[i][j]=distmat[j][i];
	      } else {
	         bioseq b2(d.Seqs[j],d.typ);
	         alignment pali(b1,b2,distmat[i][j],sub,id);
		 pali.optimal_time(distmat[i][j]);
		 cout<<d.Identif[i]<<" "<<d.Identif[j]<<" distance="<<distmat[i][j]<<endl;
	      }
	   }
       }
   }

}


double mult_alignment::median_scale(vector<vector<double> >& scalemat)
{
   // Median der Eintraege in der rechten oberen Haelfte von scalemat
   vector<double> scalevec;
   for(unsigned int i=0;i<scalemat.size();++i)
       for(unsigned int j=i+1;j<scalemat[i].size();++j)
           scalevec.push_back(scalemat[i][j]);
   sort(scalevec.begin(),scalevec.end());
   if(scalevec.size()%2==0) {
     if(scalevec.size()>0)
        return (scalevec[scalevec.size()/2-1]+scalevec[scalevec.size()/2])/2.0;
   } else
        return scalevec[scalevec.size()/2];
   return 1.0;
}

/*************************** Ausgabe ***********************/

void mult_alignment::print_phylip(ostream& outalign)
{
   printform.print_phylip(outalign);
}

void mult_alignment::print_screen(ostream& outalign)
{
   printform.print_screen(outalign);
}

void mult_alignment::print_params(ostream& pfstr)
{
   vector<double> subpars, idpars;
   unsigned int i;
   sub.paramsOf(subpars);
   id.returnparams(idpars);
   pfstr << "substitution model: ";
   for(i=0;i<subpars.size();++i)
      pfstr<<subpars[i]<<" ";
   pfstr << endl;
   pfstr << "indel model: ";
   for(i=0;i<idpars.size();++i)
      pfstr<<idpars[i]<<" ";
   pfstr << endl;
}

void mult_alignment::print_tree(ostream& trstr)
{
   told->print(trstr);
}

void mult_alignment::bericht(ostream& logfile)
{
   logfile<<"Step "<<zaehler<<":\n"
          <<"log likelihood = "<<lnL<<endl
	  <<"optimal delta = "<<coo.delta<<endl;
   print_params(logfile);
   logfile<<"tree : ";
   print_tree(logfile);
   logfile<<"alignment :\n";
   print_phylip(logfile);
   logfile<<endl;
}


/*********************** Zustand ****************************/

double mult_alignment::delta()
{
   return coo.delta;
}

bool mult_alignment::burnin()
{
   if(zaehler<coo.burn_in)
      return TRUE;
   else
      return FALSE;
}


bool mult_alignment::inrange()
{
   if(zaehler<coo.maxsteps)
      return TRUE;
   else
      return FALSE;
}


bool mult_alignment::report()
{
   if((zaehler%coo.reportafter)==0)
      return TRUE;
   else
      return FALSE;
}

double mult_alignment::likelihood()
{
   return lnL;
}

unsigned long int mult_alignment::step()
{
   return zaehler;
}

unsigned int mult_alignment::spez() const
{
   return printform.spez();
}


void mult_alignment::counterreset()
{
   if(zaehler<coo.reportafter)
      zaehler=0;
   else
      zaehler-=coo.reportafter;
}

void mult_alignment::setcounter(unsigned long int c)
{
   zaehler=c;
}

/***************** sampeln *******************************/

void mult_alignment::toldWurz()
{
   // alten Baum neu wurzeln und Likelihood berechnen
   told=told->reroot();
   lnL = complik(cosubi,told);
}

void mult_alignment::maunewton()
{
   double oldlik=lnL;
   vector<baum *> nodes; // nodes[i] und nbidx[i] kennzeichnen eine Kante
   vector<unsigned int> nbidx;    // nbidx[i] ist der Index eines Nachbarn von nodes[i]
   vector<int> zuford;   // zufaellige Ordnung der Knoten
   double newl=0.0;
   int ownch, nbch;
   baum *p;
   baum *q;
   
   ausaltmachneu(); //told nach tnew kopieren
      
   tnew->edges(nodes,nbidx); // alle Kanten werden erfasst
   rndorder(zuford,nodes.size()); // zufaellige Ordnung
   
   for(unsigned int i=0;i<zuford.size();++i) {
       // Laenge der Kante zwischen nodes[zuford[i]] und nodes[zuford[i]]->nb[nbidx[zuford[i]]].nachbar aendern
       p=nodes[zuford[i]];
       q=p->NB(nbidx[zuford[i]]);
       newl = p->shiftnodes(coo.delta*(double)(deltafaktor+MNstep)/(double)(4+MNstep),nbidx[zuford[i]]);   
       if(newl<0.0) { // Topologie aendern
          // zufaellige Kinder von p und q, die nicht p oder q sind
          ownch=p->rndchild(nbidx[zuford[i]]);
	  nbch=q->rndchild(q->whichone(p));
	  // neues Alignment der sechs Sequenzen inkl. neuen inneren Sequenzen
	  hexali(p,q,ownch,nbch);
	  //Aeste tauschen
          p->swapbr(nbidx[zuford[i]],ownch,nbch);
          // Die Kante zwischen q und seinem nbch-ten Kind
          // wird jetzt durch p und dessen ownch-ten Kind beschrieben
          keeptrack(nodes,nbidx,p,ownch,q,nbch);
       }
   }
   
   ++zaehler;

   bool nullseq=!uphill(tnew);  // ein bisschen optimieren
   if(nullseq) {
      cerr<<"WARNING: One of the internal sequences has either length 0 \nor is much longer than its longest neighbour.\n";
   }
   // Entscheidung zwischen tnew und told
   alteroderneuerbaum();
   if(zaehler==coo.burn_in && (maxdifftree>1.0-DBL_MIN && maxdifftree<1.0+DBL_MIN))
   {
       cout << "No significant improvement of the alignment during burn-in\n"
            << "Doing more burn-in steps\n";
       coo.burn_in*=2;
   }
   // justieren von delta
   wieweit[deltafaktor-1]+=fabs(oldlik-lnL);
   ++(times_visited[deltafaktor-1]);
   if(zaehler%400==0) {
   // justieren von delta
     unsigned int idx=3;  // entspricht einer Vorliebe fuer den alten Wert von delta ((3+1+MNstep)/(4+MNstep)==1)
     for(unsigned int i=0;i<wieweit.size();++i)
         if(wieweit[i]/(double)times_visited[i] > wieweit[idx]/(double)times_visited[idx])
	    idx=i;
     coo.delta*=(double)(idx+1+MNstep)/(double)(4+MNstep);
     deltafaktor=8;
     descending=TRUE;
     ++MNstep;
     for(unsigned int i=0;i<wieweit.size();++i) {
         wieweit[i]=0.0;
         times_visited[i]=0;
     }
   } else {
       if(descending) {
          --deltafaktor;
	  if(deltafaktor==1)
	     descending=FALSE;
       } else {
          ++deltafaktor;
	  if(deltafaktor==8)
	     descending=TRUE;
       }
   }
}

bool mult_alignment::trytoimprove()
{
   ausaltmachneu();
   tnew=tnew->reroot();
   uphill(tnew);
   mdataset newprintform(tnew,sub.retalf());//neues Alignment der Blaetter in ueblicher Form
   compact_subst_pattern newcosubi(newprintform);// neue Information ueber die Substitutionsmuster
   double newlik=complik(newcosubi,tnew);

   if(newlik>lnL) {
     ausneumachalt();
     printform=newprintform;
     cosubi=newcosubi;
     lnL=newlik;
     return TRUE;
   } else
     return FALSE;
}


void mult_alignment::optim_times()
{
  cout << "Optimising branch lengths\n";
  optimaltime(told);
  printform=mdataset(told,sub.retalf());//neues Alignment der Blaetter in ueblicher Form
  cosubi=compact_subst_pattern(printform);// neue Information ueber die Substitutionsmuster
  lnL=complik(cosubi,told);
}

void mult_alignment::optimaltime(baum *p)
{
    for(unsigned int i=0;i<p->nb.size();++i)
        if(!(p->nb[i].toroot)) {
	   baum *q=p->nb[i].nachbar;
	   alignment pairali(*p,*q,sub,id,coo.coolali,INT_MAX);
	   p->nb[i].length=pairali.opt_for_giv();
	   q->nb[q->whichone(p)].length=p->nb[i].length;
	   optimaltime(q);
        }
}


bool mult_alignment::nearestneighbor(unsigned int wieviele)
{
  bool retval=FALSE;
  if(told->nb.size()>0 && told->nb[0].nachbar->numleaves()>2) { // nur fuer mehr als vier Taxa sinnvoll
      ausaltmachneu();
      // zufaellige innere Kante waehlen
      vector<baum *> nodes; // nodes[i] und nbidx[i] kennzeichnen eine Kante
      vector<unsigned int> nbidx;    // nbidx[i] ist der Index eines Nachbarn von nodes[i]
      vector<int> zuford;   // zufaellige Ordnung der Knoten
      int ownch=-1, nbch=-1;
      baum *p;
      baum *q;
      tnew->internaledges(nodes,nbidx); // alle Kanten werden erfasst
      rndorder(zuford,nodes.size()); // zufaellige Ordnung
      vector<int> randint;
      for(unsigned int i=0;i<zuford.size();++i)
          if(!nodes[zuford[i]]->nb[nbidx[zuford[i]]].fix)
	     randint.push_back(zuford[i]);
      for(unsigned int brc=0; brc<randint.size() && brc<wieviele; ++brc) {
         p=nodes[randint[brc]];
         q=p->NB(nbidx[randint[brc]]);
         // fuer die vier Anlieger Distanzmatrix schaetzen
         vector<baum *> children(4);
         unsigned int i, j;
	 j=0;
	 for(i=0;i<p->nb.size();++i) {
	     if(p->nb[i].nachbar!=q) {
#ifdef DEBUG
	        if(j>=2) {
		   cerr <<"Error I in NNI routine\n";
		   exit(EXIT_FAILURE);
		}
#endif
		children[j]=p->nb[i].nachbar;
		++j;
	     }
	 }
#ifdef DEBUG
	 if(j!=2) {
	    cerr <<"Error II in NNI routine\n";
	    exit(EXIT_FAILURE);
	 }
#endif
	 for(i=0;i<q->nb.size();++i) {
	     if(q->nb[i].nachbar!=p) {
	        if(j>=4) {
		   cerr <<"Error III in NNI routine\n";
		   exit(EXIT_FAILURE);
		}
		children[j]=q->nb[i].nachbar;
		++j;
	     }
	 }
#ifdef DEBUG
	 if(j!=4) {
	    cerr <<"Error IV in NNI routine\n";
	    exit(EXIT_FAILURE);
	 }
#endif

         // Falls die Kinder keine Blaetter sind, bekommen sie neue Sequenzen
	 // ohne Beruecksichtigung von p und q
	 vector<unsigned int> enkel;
         for(unsigned int k=0;k<4;++k)
	     if(!children[k]->isLeaf()) {
                enkel.clear();
		for(unsigned int l=0;l<children[k]->nb.size();++l)
		    if(children[k]->nb[l].nachbar!=p && children[k]->nb[l].nachbar!=q)
		       enkel.push_back(l);
                midseq(children[k],enkel);
	 }

	 // paarweise Distanzen schaetzen
         double anfangszeit=1.0;
         double distmat[4][4];
         for(i=0;i<4;++i) {
             for(j=0;j<4;++j) {
               if(j==i) {
	          distmat[i][j]=0.0;
	       } else {
	         if(j<i) {
	            distmat[i][j]=distmat[j][i];
	         } else {
	            distmat[i][j]=anfangszeit;
	            alignment pali(children[i]->pro,children[j]->pro,distmat[i][j],sub,id);
		    pali.optimal_time(distmat[i][j]);
                 }
              }
	    }
	 }
         
	 // innere Kante gemaess NJ-Baum der Anlieger aendern
         bool change=FALSE;
	 if(distmat[0][2]+distmat[1][3]<distmat[0][1]+distmat[2][3] && 
	    distmat[0][2]+distmat[1][3]<distmat[0][3]+distmat[1][2]) {
	    ownch=p->whichone(children[1]);
	    nbch=q->whichone(children[2]);
	    change=TRUE;
	 } else {
	    if(distmat[0][3]+distmat[1][2]<distmat[0][1]+distmat[2][3] && 
	       distmat[0][3]+distmat[1][2]<distmat[0][2]+distmat[1][3]) {
	       ownch=p->whichone(children[1]);
	       nbch=q->whichone(children[3]);
	       change=TRUE;
	    }	 
	 }
	 retval=(retval || change);
	 if(change) {
	    cout<<"Performing nearest neighbour interchange\n";
            //Aeste tauschen
            p->swapbr(nbidx[randint[brc]],ownch,nbch);
	    keeptrack(nodes,nbidx,p,ownch,q,nbch);
	 }
	 // auf jeden Fall neue innere Sequenzen, sonst gibt's Probleme
	 vector<unsigned int> pKinder, qKinder;
         for(unsigned int i=0;i<p->nb.size();++i)
           if(p->nb[i].nachbar!=q)
	      pKinder.push_back(i);
         for(unsigned int i=0;i<q->nb.size();++i)
           if(q->nb[i].nachbar!=p)
	      qKinder.push_back(i);
	 midseq(p,pKinder);
	 midseq(q,qKinder);
	 baum *vater=NULL, *sohn=NULL;
	 if(p->nb[p->whichone(q)].toroot) {
	    vater=q; sohn=p;
	 } else {
	    vater=p; sohn=q;
	 }
	 alignment pairali(vater->pro,sohn->pro,vater->nb[vater->whichone(sohn)].length,sub,id,coo.coolali,INT_MAX);  // paarweises Alignment zum Nachbarn
	 pairali.zurueck(vater,sohn); // gegenseitige Ausrichtung wird zurueck in den Baum geschrieben
     }
  }
  if(retval) {
     ausneumachalt();
     // Likelihood neu berechnen
     printform=mdataset(told,sub.retalf());
     cosubi=compact_subst_pattern(printform);
     lnL=complik(cosubi,told);
  }
  return retval;
}


void mult_alignment::sampparams()
{
   substmodel newsub(sub);
   indel newid(id);
   proposal prop(newsub,newid,kee);
   newsub.setparams(prop.subpar);
   newid.setparams(prop.idpar);
   
   double newlik=complik(cosubi,told,newid,newsub);
   double R=randomunitintervall();
   double coun=newlik+prop.logcounter;
   double deno=lnL+prop.logdenominator;
   double decision;
   if(zaehler<=coo.burn_in)
     decision=exp(coun-deno);
   else
     decision=exp((coun-deno)*(1.0+(double)(zaehler-coo.burn_in)*betaparam)/maxdiffparam);
   if(coun>deno || R<decision) { 
     sub=newsub;
     id=newid;
     lnL=newlik;
     if(zaehler<=coo.burn_in && coun-deno>maxdiffparam && zaehler>MINDESTENS) // Spruenge am Anfang sind untypisch gross!!
         maxdiffparam=coun-deno;
   }

/*  Das folgende verhindert das Enfrieren!!! 
   // Alignment wird noch mal optimiert
   uphill(told);
   printform=mdataset(told,sub.retalf()); // Alignment der Blaetter in ueblicher Form
   cosubi=compact_subst_pattern(printform); // Information ueber die Substitutionsmuster
   lnL = complik(cosubi,told);
*/   
   if(zaehler==coo.burn_in && (maxdiffparam>1.0-DBL_MIN && maxdiffparam<1.0+DBL_MIN))
   {
       cout << "No significant improvement of the parameters during burn-in\n"
            << "Doing more burn-in steps\n";
       coo.burn_in*=2;
   }
   if(zaehler==coo.burn_in) {
      double numtax=(double)(printform.spez());
      double alile=(double)(printform.laen(0));
      betaparam = coo.cparam/((1.0-coo.alphaparam)*numtax-lnL*coo.alphaparam/alile);
   }
}

/************* private Funktionen     *****************************/


/*********   Konstruktion und Optimierung   ******************/

void mult_alignment::construct_align(baum *p)
{
   vector<unsigned int> Kinder;
   
   if(p->isRoot() || !p->isLeaf()) { // "Wurzelblatt" oder innerer Knoten
      for(unsigned int i=0;i<p->nb.size();++i)
         if(!(p->nb[i].toroot)) {
            construct_align(p->nb[i].nachbar);
	    Kinder.push_back(i);
	 }
         if(p->nb.size()==3)     // innerer Knoten
	     midseq(p,Kinder);  // ermittelt Profil an innerem Knoten
	 else {
	    if(p->nb.size()==1) {   // "Wurzelblatt"
	       alignment pairali(p->pro,p->nb[0].nachbar->pro,p->nb[0].length,sub,id,coo.coolali,INT_MAX);  // paarweises Alignment zum Nachbarn
	       pairali.zurueck(p,p->nb[0].nachbar); // gegenseitige Ausrichtung wird zurueck in den Baum geschrieben
            } else {
               cerr << "Tree in function construct_align is not bifurcating!\n";
	       exit(EXIT_FAILURE);
	    }
         }
   }
}

void mult_alignment::midseq(baum* p, vector<unsigned int>& Kinder)
{
   if(Kinder.size()!=2) {
      cerr << "Tree in function midseq is not bifurcating!\n";
      exit(EXIT_FAILURE);
   }
   // nexttriplet betrachtet das rechte Kind als ancestral.
   // Deshalb wird's so gemacht:
   alignment pairali(p->nb[Kinder[1]].nachbar->pro,p->nb[Kinder[0]].nachbar->pro,
                     p->nb[Kinder[0]].length+p->nb[Kinder[1]].length,sub,id,coo.coolali,INT_MAX);
   id.fillT(p->nb[Kinder[0]].length,p->nb[Kinder[1]].length);  // Zeiten des Indelmodells setzen
   vector<zustand> zuse;
   int i=-1,j=-1;
   zuse.push_back(pairali.iter_zust(&i,&j));
   while(zuse.back()!=ENDE)
       zuse.push_back(pairali.iter_zust(&i,&j));
   vector<triplet_zus> dreiseq;
   unsigned int k=0;
   dreiseq.push_back(tSTART);	     
   while(dreiseq.back()!=tENDE && k<zuse.size()) {
      switch(dreiseq.back()) {
          case tXYZ :
	  case tXY_ :
	  case tX_Z :
	  case tSTART :
	  case t_Y_1:
	  case t_Y_2:
	  case t__Z :
             ++k;
	     break;
          case tX__ :
	     break;
	  default :
	     cerr << "Impossible state in function midseq\n";
	     exit(EXIT_FAILURE);
      }
      dreiseq.push_back(id.nexttriplet(dreiseq.back(),zuse[k]));		     
   }
   if(dreiseq.back()!=tENDE) {
      cerr << "Error in function midseq\n";
      exit(EXIT_FAILURE);
   }
      
   write_midseq_to_tree(dreiseq,p,Kinder);  // Profil an Knoten p und Ausrichtung gegen die Kinder zurueckschreiben
}

bool mult_alignment::uphill(baum *tree)            // optimieren des Alignments
{
   return uphill(tree,sub,id);
}

bool mult_alignment::uphill(baum *tree, substmodel& submod, indel& indmod)            // optimieren des Alignments
{
   bool retval=TRUE;
   if(!tree->isRoot()) {
      cerr << "Lost root\n";
      exit(EXIT_FAILURE);
   }

   vector<baum *> int_nodes; // interne Knoten
   vector<baum *> nodes; // nodes[i] und nbidx[i] kennzeichnen eine Kante
   vector<unsigned int> nbidx;    // nbidx[i] ist der Index eines Nachbarn von nodes[i]
   vector<int> zuford;   // zufaellige Ordnung der Knoten
   unsigned int i, idx;
   
   // alle inneren Knoten und alle Kanten erfassen
   tree->edges(nodes,nbidx); // alle Kanten werden erfasst
   tree->intnodes(int_nodes); // alle inneren Knoten werden erfasst
   // zufaellige Reihenfolge der Knoten und Kanten erzeugen
   rndorder(zuford,nodes.size()+int_nodes.size());
   
   // fuer alle Kanten und inneren Knoten
   for(i=0;i<zuford.size();++i)
   {
      if(zuford[i]<(int)int_nodes.size()) {
         // falls innerer Knoten: neue innere Sequenz und neues Viereralignment sampeln

	 // Was gibt's fuer Nachbarn
	 vector<unsigned int> Kinder;
	 familie(Kinder,int_nodes[zuford[i]]);
	 // erzeuge mittlere Sequenz mit zugehoerigen Alignments
         samp_centseq(int_nodes[zuford[i]],Kinder,submod,indmod);
	 if(int_nodes[zuford[i]]->pro.laen()==0) // erzeugte Sequenz sollte nicht die Laenge 0 haben
	    return FALSE;
	 // die Sequenz sollte aber auch nicht zu lang sein
	 retval=(int_nodes[zuford[i]]->pro.laen()<MAXMULT*int_nodes[zuford[i]]->nb[Kinder[0]].nachbar->pro.laen());
	 for(unsigned int j=1;j<Kinder.size();++j)
	    retval=(retval || 
	            int_nodes[zuford[i]]->pro.laen()<MAXMULT*int_nodes[zuford[i]]->nb[Kinder[j]].nachbar->pro.laen());
	 if(!retval)
	    return retval;
      } else {
         // falls Kante: neues Alignment sampeln
	 idx=zuford[i]-int_nodes.size();
         alignment pairali(*(nodes[idx]),*(nodes[idx]->nb[nbidx[idx]].nachbar),submod,indmod,coo.coolali,INT_MAX);
	 pairali.teilsample(coo.resamprange);
         pairali.zurueck(nodes[idx],nodes[idx]->nb[nbidx[idx]].nachbar);
      }
   }
   
   return retval;
}


void mult_alignment::samp_centseq(baum *b, vector<unsigned int>& Kinder, substmodel& submod, indel& indmod)
{
  vector<triplet_zus> tri;
  vector<quart_zus> quart;
  
  if(b==NULL || Kinder.size()!=3) {
     cerr << "Unreasonable arguments to samp_midseq!\n";
     exit(EXIT_FAILURE);
  }
  quartali(quart,b,Kinder);
  quart_to_tri(quart,tri);
  
  indmod.fillQ(b->nb[Kinder[0]].length,b->nb[Kinder[1]].length,b->nb[Kinder[2]].length);
  int i=1;  // tri[0] ist tSTART und wird spaeter nicht gebraucht
  quart.clear();
  quart.push_back(qSTART);
  while(quart.back()!=qENDE && i<(int)tri.size()) {
         quart.push_back(id.nextquartet(quart.back(),tri[i]));
         switch(quart.back()) {
            case qWXYZ :
	    case qWXY_ :
	    case qWX_Z :
	    case q_XYZ :
	    case qWX__ :
            case q_XY_ :
	    case q_X_Z :
	    case qW___ :
	    case q__Y_1 :
	    case q__Y_2 :
	    case q__Y_3 :
	    case q__Y_4 :
	    case q___Z1 :
	    case q___Z2 :
	    case qENDE :
	      ++i;
	      break;
	    case q_X__ :
              break;
            default :
	      cerr << "Unreasonable state in samp_centseq!\n";
	      exit(EXIT_FAILURE);
         }
  }
  if(quart.back()!=qENDE) {
        cerr << "Error in function samp_centseq!\n";
        exit(EXIT_FAILURE);
  }
  write_centseq(quart,b,Kinder,submod);
}


void mult_alignment::central_pair(mercedes_stern& m1, mercedes_stern& m2)
{
    vector<hex_zus> allesequs;
    vector<quart_zus> aussen;
    mercedes_stern *anc=NULL, *rec=NULL;  // urspruengliches Verwandtschaftsverhaeltnis (rec ist ein Kind von anc)
    swapflag swp=which_pair_swaps(m1,m2,anc,rec);// Welcher Mercedesstern ist der aeltere?
                                                 // u. U. die Reihenfolge der Verwandten des aelteren Sterns aendern
    zweimalvier(aussen,*anc,*rec,swp);
    fillH(*anc,*rec,swp);
    allesequs.push_back(hSTART);
    unsigned int i=1;  // qSTART ueberspringen
    while(allesequs.back()!=hENDE) {
          if(i>=aussen.size()) {
             cerr << "Error in function central_pair\n";
	     exit(EXIT_FAILURE);
          }
          allesequs.push_back(id.nexthextet(allesequs.back(),aussen[i]));
          switch(allesequs.back()) {
             case hUVWXYZ :
	     case hUV_XYZ :
	     case hUVWX_Z :
	     case hUV_X_Z :
             case hUVWXY_ :
	     case hUV_XY_ :
	     case hUVWX__ :
	     case hUV_X__ :
	     case h_VWXYZ :
	     case h_V_XYZ :
	     case h_VWX_Z :
	     case h_V_X_Z :
	     case h_VWXY_ :
	     case h_V_XY_ :
	     case h_VWX__ :
	     case hUV____ :
	     case hUVW___ :
	     case h_VW___ :
	     case h___XYZ :
	     case h___X_Z :
	     case h___XY_ :
	     case hU_____ :
	     case hW1 :
	     case hW2 :
	     case hW3 :
	     case hW4 :
	     case hW5 :
	     case hY1 :
	     case hY2 :
	     case h_____Z :
               ++i;
	       break;
	     case h_V_X__ :
	     case h_V____ :
	     case h___X__ :
	       break;
	     case hENDE :
	       if(i!=aussen.size()-1) {
	          cerr << "Premature end in central_pair\n";
	          exit(EXIT_FAILURE);
	       }
	       break;
	     default :
	       cerr << "Unreasonable state in function central_pair\n";
	       exit(EXIT_FAILURE);
          }
    }
    write_centpair(allesequs,*anc,*rec,swp);
}


bool mult_alignment::alteroderneuerbaum()
{
  mdataset newprintform(tnew,sub.retalf());//neues Alignment der Blaetter in ueblicher Form
  compact_subst_pattern newcosubi(newprintform);// neue Information ueber die Substitutionsmuster
  double newlik=complik(newcosubi,tnew);
  double R=randomunitintervall();
  double decision;
  bool retval=FALSE;
  if(zaehler<=coo.burn_in)
     decision=exp(newlik-lnL);
  else
     decision=exp((newlik-lnL)*(1.0+(double)(zaehler-coo.burn_in)*betatree)/maxdifftree);

  if(newlik>lnL || R<decision) {
     retval=TRUE;
     ausneumachalt();
     printform=newprintform;
     cosubi=newcosubi;
     if(zaehler<=coo.burn_in && newlik-lnL>maxdifftree && zaehler>MINDESTENS) // Spruenge am Anfang sind untypisch gross!!
	  maxdifftree=newlik-lnL;
     lnL=newlik;
  }
  if(zaehler==coo.burn_in) {
      double numtax=(double)(printform.spez());
      double alile=(double)(printform.laen(0));
      betatree=coo.ctree/((1.0-coo.alphatree)*numtax-lnL*coo.alphatree/alile);
  }
  
  return retval;
}


/*********   berechnen der Likelihood   ******************/

double mult_alignment::complik(compact_subst_pattern& csb, baum *tree) 
// logLikelihood des Alignments
{
   return complik(csb,tree,id,sub);
}

double mult_alignment::complik(compact_subst_pattern& csb, baum *tree, indel& newid, substmodel& newsub) 
// logLikelihood des Alignments fuer die angegebenen Indel- und Substitutionsmodelle
{
   return indlik(tree,newid)+sublik(csb,tree,newsub);
}


double mult_alignment::indlik(baum *tree, indel& newid) 
// Insertions-Deletions-Summand der logLikelihood des Alignments fuer das Indelmodell newid
{
   double loglik=0.0;
   unsigned int i;
   baum *nachbar;
   int tidx;
   vector<int> idces(2);
   zustand z1, z2;
   
   for(i=0;i<tree->nb.size();++i)
      if(!tree->nb[i].toroot) {
         idces[0]=idces[1]=-1;
	 nachbar=tree->nb[i].nachbar;
	 tidx=nachbar->whichone(tree);
	 newid.fillP(tree->nb[i].length);
         z1=iter_zust(tree->nb[i].ausrichter,nachbar->nb[tidx].ausrichter,idces);
         do {
           z2=iter_zust(tree->nb[i].ausrichter,nachbar->nb[tidx].ausrichter,idces);
	   loglik+=newid.retlnprob(z1,z2);
	   z1=z2;
         }while(z1!=ENDE);
	 if(!tree->isRoot())
	   loglik-=newid.logprob_of_seqlen(tree->pro.size());
	 loglik+=indlik(nachbar,newid);
   }
   return loglik;
}

double mult_alignment::sublik(compact_subst_pattern& csb, baum *tree, substmodel& newsub) 
// Substitutions-Summand der logLikelihood des Alignments fuer das Subst.modell newsub
{
   double logsum=0.0;
   vector<vector<double> > loglik;
   vector<double> finally;
   baum *t=NULL;
   unsigned int i, j;
   
   loglik.resize(1);
   loglik[0].resize(csb.muster.size());
   for(i=0;i<loglik[0].size();++i)
       loglik[0][i]=0.0;
   for(i=loglik[0].size()/2;i>1;i/=2){
       vector<double> newback(i,0.0);
       loglik.push_back(newback); 
   }

   for(i=0;i<csb.muster.size();++i) {
      if(i==0) {
        t = tree->teilbaum(csb.ids,csb.muster[i]);
      } else {
          if(!patternequivalent(csb.muster[i-1],csb.muster[i])) {
	     delete(t);
             t = tree->teilbaum(csb.ids,csb.muster[i]);
	  }
      }
      loglik[0][i]+=(double)(csb.anzmus[i])*logprobofpattern(csb.muster[i],t,newsub);
   }
   for(i=1;i<loglik.size();++i)
     for(j=0;j<loglik[i-1].size();++j)
       loglik[i][j%loglik[i].size()]+=loglik[i-1][j];

  finally=loglik.back();
  for(i=0;i<finally.size();++i)
      logsum+=finally[i];
   if(t!=NULL)
      delete(t);
   return logsum;
}



double mult_alignment::logprobofpattern(vector<char>& pattern, baum *tree, substmodel& newsub)
{
    int i;
    alphsize alf=newsub.retalf();
    vector<double> logliks, frtilo;  //logliks wird in childliks initialisiert

    childliks(pattern,tree,newsub,logliks,alf);
    if(tree->isLeaf()) {
       i=whichbase(pattern[tree->id.number],alf);
       frtilo.push_back(newsub.logfreqofstate(i)+logliks[i]);
    } else {
       for(i=0;i<alf;++i)  
         frtilo.push_back(newsub.logfreqofstate(i)+logliks[i]);
    }

    return addlog(frtilo);
}


void mult_alignment::childliks(vector<char>& pattern, baum *tree, substmodel& newsub, vector<double>& logliks, int alf)
{
    vector<double> chlogliks, frtilo;
    unsigned int i, j, k;
    int leafstate;
    
    if(tree->isLeaf())
       logliks.resize(alftologalf(alf));
    else
       logliks.resize(alf);
    for(i=0;i<logliks.size();++i)
        logliks[i]=-0.0;
        
    for(i=0;i<tree->nb.size();++i) 
        if(!tree->nb[i].toroot) {  // fuer jeden Nachbarn, das nicht Richtung Wurzel liegt
           newsub.settime(tree->nb[i].length); // Laenge des Astes setzen
	   if(!tree->nb[i].nachbar->isLeaf()) { // Rekursion falls Nachbar kein Blatt 
               childliks(pattern,tree->nb[i].nachbar,newsub,chlogliks,alf);
               for(j=0;j<logliks.size();++j) {
                   frtilo.clear();
	           for(k=0;k<chlogliks.size();++k)
	               frtilo.push_back(newsub.logProb_ij(j,k)+chlogliks[k]);
                   logliks[j]+=addlog(frtilo);
	       }
	   } else {  // Nachbar ist ein Blatt
	       leafstate=whichbase(pattern[tree->nb[i].nachbar->id.number],alf); // Welche Base steht am Nachbarblatt?
	       for(j=0;j<logliks.size();++j)
	           logliks[j]+=newsub.logProb_ij(j,leafstate);
	   }
        }
}


/********* zurueckschreiben in den Baum   ******************/

void mult_alignment::write_midseq_to_tree(vector<triplet_zus>& dreiseq, baum *p, vector<unsigned int>& Kinder)
{
   if(p->pro.size()!=0) {
      //cerr << "Central sequence already exists. It is now destroyed.\n";
      p->pro.clear();
   }
   if(p->nb[Kinder[0]].ausrichter.size()!=0) {
      //cerr << "Alignment already exists. It is now destroyed.\n";
      p->nb[Kinder[0]].ausrichter.clear();
   }
   if(p->nb[Kinder[1]].ausrichter.size()!=0) {
      //cerr << "Alignment already exists. It is now destroyed.\n";
      p->nb[Kinder[1]].ausrichter.clear();
   }
   int yidx = p->nb[Kinder[0]].nachbar->whichone(p); 
   if(p->nb[Kinder[0]].nachbar->nb[yidx].ausrichter.size()!=0) {
      //cerr << "Alignment already exists. It is now destroyed.\n";
      p->nb[Kinder[0]].nachbar->nb[yidx].ausrichter.clear();
   }
   int zidx = p->nb[Kinder[1]].nachbar->whichone(p); 
   if(p->nb[Kinder[1]].nachbar->nb[zidx].ausrichter.size()!=0) {
      //cerr << "Alignment already exists. It is now destroyed.\n";
      p->nb[Kinder[1]].nachbar->nb[zidx].ausrichter.clear();
   }
   int proy, proz;
   int lx=-1, ly=-1, lz=-1;
   double txy, txz;
   txy=p->nb[Kinder[0]].length;
   txz=p->nb[Kinder[1]].length;
   substmodel smy(sub), smz(sub);
   smy.settime(txy);
   smz.settime(txz);
   for(unsigned int i=0;i<dreiseq.size();++i) {
       switch(dreiseq[i]) {
          case tSTART :
	     if(i!=0) {
	        cerr << "Impossible state in write_midseq_to_tree\n";
		exit(EXIT_FAILURE);
	     }
	     break;
          case tENDE :
	     if(i!=dreiseq.size()-1) {
	        cerr << "Impossible state in write_midseq_to_tree\n";
		exit(EXIT_FAILURE);
	     }
	     break;
	  case tXYZ :
	     ++lx;
	     ++ly;
	     ++lz;
	     p->nb[Kinder[0]].ausrichter.push_back(ly);
	     p->nb[Kinder[1]].ausrichter.push_back(lz);
	     p->nb[Kinder[0]].nachbar->nb[yidx].ausrichter.push_back(lx);
	     p->nb[Kinder[1]].nachbar->nb[zidx].ausrichter.push_back(lx);
             proy=p->nb[Kinder[0]].nachbar->pro[ly];
             proz=p->nb[Kinder[1]].nachbar->pro[lz];
	     p->pro.push_back(infer_base(proy,smy,proz,smz)); 
	     break;
	  case tXY_ :
	     ++lx;
	     ++ly;
	     p->nb[Kinder[0]].ausrichter.push_back(ly);
	     p->nb[Kinder[1]].ausrichter.push_back(lz);
	     p->nb[Kinder[0]].nachbar->nb[yidx].ausrichter.push_back(lx);
             proy=p->nb[Kinder[0]].nachbar->pro[ly];
	     p->pro.push_back(infer_base(proy,smy)); 
	     break;
	  case tX_Z :
	     ++lx;
	     ++lz;
	     p->nb[Kinder[0]].ausrichter.push_back(ly);
	     p->nb[Kinder[1]].ausrichter.push_back(lz);
	     p->nb[Kinder[1]].nachbar->nb[zidx].ausrichter.push_back(lx);
             proz=p->nb[Kinder[1]].nachbar->pro[lz];
	     p->pro.push_back(infer_base(proz,smz)); 
	     break;
	  case t__Z :
	     ++lz;
	     p->nb[Kinder[1]].nachbar->nb[zidx].ausrichter.push_back(lx);
	     break;
	  case t_Y_1 :
	  case t_Y_2 :
	     ++ly;
	     p->nb[Kinder[0]].nachbar->nb[yidx].ausrichter.push_back(lx);
	     break;
          case tX__ :
	     ++lx;
	     p->nb[Kinder[0]].ausrichter.push_back(ly);
	     p->nb[Kinder[1]].ausrichter.push_back(lz);
	     p->pro.push_back(sub.choosebase()); 
	     break; 
          default :
	     cerr << "Impossible state in write_midseq_to_tree\n";
	     exit(EXIT_FAILURE);
       }
   }
   if(lx!=(int)p->pro.size()-1 
      || ly!=(int)p->nb[Kinder[0]].nachbar->pro.size()-1 
      || lz!=(int)p->nb[Kinder[1]].nachbar->pro.size()-1) {
         cerr << "Error in function write_midseq_to_tree:\n"
              <<(int)p->pro.size()-1-lx<<" "
              <<(int)p->nb[Kinder[0]].nachbar->pro.size()-1-ly<<" "
	      <<(int)p->nb[Kinder[1]].nachbar->pro.size()-1-lz<<endl;
	 exit(EXIT_FAILURE);
   }
}


void mult_alignment::write_centseq(vector<quart_zus>& quart, baum *b, vector<unsigned int>& Kinder, substmodel& submod)
{
   if(Kinder.size()!=3 || b->nb.size()!=3) {
      cerr << "Strange arguments to function writecentseq!\n";
      exit(EXIT_FAILURE);
   }
   int widx = b->nb[Kinder[0]].nachbar->whichone(b);
   int yidx = b->nb[Kinder[1]].nachbar->whichone(b);
   int zidx = b->nb[Kinder[2]].nachbar->whichone(b);
   b->nb[Kinder[0]].ausrichter.clear();
   b->nb[Kinder[0]].nachbar->nb[widx].ausrichter.clear();
   b->nb[Kinder[1]].ausrichter.clear();
   b->nb[Kinder[1]].nachbar->nb[yidx].ausrichter.clear();
   b->nb[Kinder[2]].ausrichter.clear();
   b->nb[Kinder[2]].nachbar->nb[zidx].ausrichter.clear();
   b->pro.clear();
   int prow, proy, proz;
   int lw=-1, lx=-1, ly=-1, lz=-1;
   double twx, txy, txz;
   twx=b->nb[Kinder[0]].length;
   txy=b->nb[Kinder[1]].length;
   txz=b->nb[Kinder[2]].length;
   substmodel smw(submod), smy(submod), smz(submod);
   smw.settime(twx);
   smy.settime(txy);
   smz.settime(txz);
   for(unsigned int i=0;i<quart.size();i++) {
       switch(quart[i]) {
          case qSTART :
	     if(i!=0) {
	        cerr << "Impossible state in write_centseq\n";
		exit(EXIT_FAILURE);
	     }
	     break;
          case qENDE :
	     if(i!=quart.size()-1) {
	        cerr << "Impossible state in write_centseq\n";
		exit(EXIT_FAILURE);
	     }
	     break;
          case qWXYZ :
             ++lw;++lx;++ly;++lz;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[0]].nachbar->nb[widx].ausrichter.push_back(lx);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[1]].nachbar->nb[yidx].ausrichter.push_back(lx);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     b->nb[Kinder[2]].nachbar->nb[zidx].ausrichter.push_back(lx);
	     prow = b->nb[Kinder[0]].nachbar->pro[lw];
	     proy = b->nb[Kinder[1]].nachbar->pro[ly];
	     proz = b->nb[Kinder[2]].nachbar->pro[lz];
	     b->pro.push_back(infer_base(prow,smw,proy,smy,proz,smz));
             break;
	  case qWXY_ :
             ++lw;++lx;++ly;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[0]].nachbar->nb[widx].ausrichter.push_back(lx);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[1]].nachbar->nb[yidx].ausrichter.push_back(lx);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     prow = b->nb[Kinder[0]].nachbar->pro[lw];
	     proy = b->nb[Kinder[1]].nachbar->pro[ly];
	     b->pro.push_back(infer_base(prow,smw,proy,smy));
             break;
	  case qWX_Z :
             ++lw;++lx;++lz;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[0]].nachbar->nb[widx].ausrichter.push_back(lx);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     b->nb[Kinder[2]].nachbar->nb[zidx].ausrichter.push_back(lx);
	     prow = b->nb[Kinder[0]].nachbar->pro[lw];
	     proz = b->nb[Kinder[2]].nachbar->pro[lz];
	     b->pro.push_back(infer_base(prow,smw,proz,smz));
             break;
	  case q_XYZ :
             ++lx;++ly;++lz;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[1]].nachbar->nb[yidx].ausrichter.push_back(lx);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     b->nb[Kinder[2]].nachbar->nb[zidx].ausrichter.push_back(lx);
	     proy = b->nb[Kinder[1]].nachbar->pro[ly];
	     proz = b->nb[Kinder[2]].nachbar->pro[lz];
	     b->pro.push_back(infer_base(proy,smy,proz,smz));
             break;
	  case qWX__ :
             ++lw;++lx;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[0]].nachbar->nb[widx].ausrichter.push_back(lx);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     prow = b->nb[Kinder[0]].nachbar->pro[lw];
	     b->pro.push_back(infer_base(prow,smw));
             break;
          case q_XY_ :
             ++lx;++ly;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[1]].nachbar->nb[yidx].ausrichter.push_back(lx);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     proy = b->nb[Kinder[1]].nachbar->pro[ly];
	     b->pro.push_back(infer_base(proy,smy));
             break;
	  case q_X_Z :
             ++lx;++lz;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     b->nb[Kinder[2]].nachbar->nb[zidx].ausrichter.push_back(lx);
	     proz = b->nb[Kinder[2]].nachbar->pro[lz];
	     b->pro.push_back(infer_base(proz,smz));
             break;
	  case qW___ :
             ++lw;
	     b->nb[Kinder[0]].nachbar->nb[widx].ausrichter.push_back(lx);
             break;
	  case q__Y_1 :
	  case q__Y_2 :
	  case q__Y_3 :
	  case q__Y_4 :
             ++ly;
	     b->nb[Kinder[1]].nachbar->nb[yidx].ausrichter.push_back(lx);
             break;
	  case q___Z1 :
	  case q___Z2 :
             ++lz;
	     b->nb[Kinder[2]].nachbar->nb[zidx].ausrichter.push_back(lx);
             break;
	  case q_X__ :
             ++lx;
	     b->nb[Kinder[0]].ausrichter.push_back(lw);
	     b->nb[Kinder[1]].ausrichter.push_back(ly);
	     b->nb[Kinder[2]].ausrichter.push_back(lz);
	     b->pro.push_back(sub.choosebase());
             break;
          default :
	     cerr << "Impossible state in write_centseq\n";
	     exit(EXIT_FAILURE);
       }
   }

}

void mult_alignment::write_centpair(vector<hex_zus>& allesequs, mercedes_stern& anc, mercedes_stern& rec, swapflag swp)
{
   if(rec.tree->parent()!=anc.tree) {
      cerr << "anc is not the parent of rec in function fillH\n";
      exit(EXIT_FAILURE);
   }

   baum *Ubaum, *Vbaum, *Wbaum, *Xbaum, *Ybaum, *Zbaum;
   int u_idx_v, v_idx_u, v_idx_w, w_idx_v, v_idx_x, x_idx_v, x_idx_y, y_idx_x, x_idx_z, z_idx_x;
   int lu, lv, lw, lx, ly, lz;
   double tuv, tvw, tvx, txy, txz;
   int profu, profv, profw, profx, profy, profz;
   
   switch(swp) {
      case WggY :
	// aus V wird X, aus X wird V 
        Vbaum=rec.tree;
	Xbaum=anc.tree;
	v_idx_x=Vbaum->whichone(Xbaum);
	x_idx_v=Xbaum->whichone(Vbaum);
	tvx=Vbaum->nb[v_idx_x].length;
        // altes U bleibt U
	Ubaum=anc.tree->nb[anc.ordered_children[0]].nachbar;
	u_idx_v=Ubaum->whichone(anc.tree);
	tuv=Ubaum->nb[u_idx_v].length;
	// altes Y bleibt Y
	Ybaum=rec.tree->nb[rec.ordered_children[1]].nachbar;
	y_idx_x=Ybaum->whichone(rec.tree);
	txy=Ybaum->nb[y_idx_x].length;
	// U sitzt jetzt, wo vorher Y sass und umgekehrt
	v_idx_u=Vbaum->whichone(Ybaum);
	x_idx_y=Xbaum->whichone(Ubaum);
	
	// aus Z wird W
	Wbaum=rec.tree->nb[rec.ordered_children[2]].nachbar;
	w_idx_v=Wbaum->whichone(rec.tree);
	tvw=Wbaum->nb[w_idx_v].length;
	// aus W wird Z
	Zbaum=anc.tree->nb[anc.ordered_children[1]].nachbar;
	if(Zbaum==rec.tree)
	   Zbaum=anc.tree->nb[anc.ordered_children[2]].nachbar;
	z_idx_x=Zbaum->whichone(anc.tree);
	txz=Zbaum->nb[z_idx_x].length; 
	
	// Z und W bleiben sitzen
	v_idx_w=Vbaum->whichone(Wbaum);
	x_idx_z=Xbaum->whichone(Zbaum);
        break;
      case WggZ :
	// aus V wird X, aus X wird V 
        Vbaum=rec.tree;
	Xbaum=anc.tree;
	v_idx_x=Vbaum->whichone(Xbaum);
	x_idx_v=Xbaum->whichone(Vbaum);
	tvx=Vbaum->nb[v_idx_x].length;
        // altes U bleibt U
	Ubaum=anc.tree->nb[anc.ordered_children[0]].nachbar;
	u_idx_v=Ubaum->whichone(anc.tree);
	tuv=Ubaum->nb[u_idx_v].length;
	// altes Z bleibt Z
	Zbaum=rec.tree->nb[rec.ordered_children[2]].nachbar;
	z_idx_x=Zbaum->whichone(rec.tree);
	txz=Zbaum->nb[z_idx_x].length;
	// U sitzt jetzt, wo vorher Z sass und umgekehrt
	v_idx_u=Vbaum->whichone(Zbaum);
	x_idx_z=Xbaum->whichone(Ubaum);

	// aus Y wird W
	Wbaum=rec.tree->nb[rec.ordered_children[1]].nachbar;
	w_idx_v=Wbaum->whichone(rec.tree);
	tvw=Wbaum->nb[w_idx_v].length;
	// aus W wird Y
	Ybaum=anc.tree->nb[anc.ordered_children[1]].nachbar;
	if(Ybaum==rec.tree)
	   Ybaum=anc.tree->nb[anc.ordered_children[2]].nachbar;
	y_idx_x=Ybaum->whichone(anc.tree);
	txy=Ybaum->nb[y_idx_x].length; 
	
	// Y und W bleiben sitzen
	v_idx_w=Vbaum->whichone(Wbaum);
	x_idx_y=Xbaum->whichone(Ybaum);
        break;
      case XggY :
	// aus V bleibt V, aus X bleibt X
        Vbaum=anc.tree;
	Xbaum=rec.tree;
	v_idx_x=Vbaum->whichone(Xbaum);
	x_idx_v=Xbaum->whichone(Vbaum);
	tvx=Vbaum->nb[v_idx_x].length;
        // altes U bleibt U
	Ubaum=anc.tree->nb[anc.ordered_children[0]].nachbar;
	u_idx_v=Ubaum->whichone(anc.tree);
	tuv=Ubaum->nb[u_idx_v].length;
        // altes Z bleibt Z
	Zbaum=rec.tree->nb[rec.ordered_children[2]].nachbar;
	z_idx_x=Zbaum->whichone(rec.tree);
	txz=Zbaum->nb[z_idx_x].length;
	// U und Z bleiben sitzen
        v_idx_u=Vbaum->whichone(Ubaum);
        x_idx_z=Xbaum->whichone(Zbaum);

	// aus Y wird W
	Wbaum=rec.tree->nb[rec.ordered_children[1]].nachbar;
	w_idx_v=Wbaum->whichone(rec.tree);
	tvw=Wbaum->nb[w_idx_v].length;
	// aus W wird Y
	Ybaum=anc.tree->nb[anc.ordered_children[1]].nachbar;
	if(Ybaum==rec.tree)
	   Ybaum=anc.tree->nb[anc.ordered_children[2]].nachbar;
	y_idx_x=Ybaum->whichone(anc.tree);
	txy=Ybaum->nb[y_idx_x].length; 
	
	// Y sitzt, wo W sass und umgekehrt
	v_idx_w=Vbaum->whichone(Ybaum);
        x_idx_y=Xbaum->whichone(Wbaum);
        break;
      case XggZ :
	// aus V bleibt V, aus X bleibt X
        Vbaum=anc.tree;
	Xbaum=rec.tree;
	v_idx_x=Vbaum->whichone(Xbaum);
	x_idx_v=Xbaum->whichone(Vbaum);
	tvx=Vbaum->nb[v_idx_x].length;
        // altes U bleibt U
	Ubaum=anc.tree->nb[anc.ordered_children[0]].nachbar;
	u_idx_v=Ubaum->whichone(anc.tree);
	tuv=Ubaum->nb[u_idx_v].length;
	// altes Y bleibt Y
	Ybaum=rec.tree->nb[rec.ordered_children[1]].nachbar;
	y_idx_x=Ybaum->whichone(rec.tree);
	txy=Ybaum->nb[y_idx_x].length;
	// U und Y bleiben sitzen
        v_idx_u=Vbaum->whichone(Ubaum);
        x_idx_y=Xbaum->whichone(Ybaum);

	// aus Z wird W
	Wbaum=rec.tree->nb[rec.ordered_children[2]].nachbar;
	w_idx_v=Wbaum->whichone(rec.tree);
	tvw=Wbaum->nb[w_idx_v].length;
	// aus W wird Z
	Zbaum=anc.tree->nb[anc.ordered_children[1]].nachbar;
	if(Zbaum==rec.tree)
	   Zbaum=anc.tree->nb[anc.ordered_children[2]].nachbar;
	z_idx_x=Zbaum->whichone(anc.tree);
	txz=Zbaum->nb[z_idx_x].length; 
	
	// Z sitzt, wo W sass und umgekehrt
	v_idx_w=Vbaum->whichone(Zbaum);
        x_idx_z=Xbaum->whichone(Wbaum);
        break;
      default :
        cerr << "Impossible value of swp in write_centpair\n";
        exit(EXIT_FAILURE);
   }
   lu=lv=lw=lx=ly=lz=-1;
   substmodel smu(sub);
   smu.settime(tuv);
   substmodel smw(sub);
   smw.settime(tvw);
   substmodel smx(sub);
   smx.settime(tvx);
   substmodel smy(sub);
   smy.settime(txy);
   substmodel smz(sub);
   smz.settime(txz);
   // alles klar machen
   Ubaum->nb[u_idx_v].ausrichter.clear();
   Vbaum->nb[v_idx_u].ausrichter.clear();
   Vbaum->nb[v_idx_w].ausrichter.clear();
   Vbaum->nb[v_idx_x].ausrichter.clear();
   Wbaum->nb[w_idx_v].ausrichter.clear();
   Xbaum->nb[x_idx_v].ausrichter.clear();
   Xbaum->nb[x_idx_y].ausrichter.clear();
   Xbaum->nb[x_idx_z].ausrichter.clear();
   Ybaum->nb[y_idx_x].ausrichter.clear();
   Zbaum->nb[z_idx_x].ausrichter.clear();
   Vbaum->pro.clear();
   Xbaum->pro.clear();
   for(unsigned int hexidx=0;hexidx<allesequs.size();++hexidx) {
       switch(allesequs[hexidx]) {
           case hUVWXYZ :
	      ++lu;++lv;++lw;++lx;++ly;++lz;
              profu=Ubaum->pro[lu];
	      profw=Wbaum->pro[lw];
	      profy=Ybaum->pro[ly];
	      profz=Zbaum->pro[lz];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_six(profv,profx,smx,profu,smu,profw,smw,profy,smy,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUV_XYZ :
	      ++lu;++lv;++lx;++ly;++lz;
              profu=Ubaum->pro[lu];
	      profy=Ybaum->pro[ly];
	      profz=Zbaum->pro[lz];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_five(profv,profx,smx,profu,smu,profy,smy,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUVWX_Z :
	      ++lu;++lv;++lw;++lx;++lz;
              profu=Ubaum->pro[lu];
	      profw=Wbaum->pro[lw];
	      profz=Zbaum->pro[lz];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_five(profx,profv,smx,profz,smz,profw,smw,profu,smu);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUV_X_Z :
	      ++lu;++lv;++lx;++lz;
              profu=Ubaum->pro[lu];
	      profz=Zbaum->pro[lz];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_line(profv,profx,smx,profu,smu,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
           case hUVWXY_ :
	      ++lu;++lv;++lw;++lx;++ly;
              profu=Ubaum->pro[lu];
	      profw=Wbaum->pro[lw];
	      profy=Ybaum->pro[ly];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      infer_five(profx,profv,smx,profy,smy,profw,smw,profu,smu);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUV_XY_ :
	      ++lu;++lv;++lx;++ly;
              profu=Ubaum->pro[lu];
	      profy=Ybaum->pro[ly];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      infer_line(profv,profx,smx,profu,smu,profy,smy);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUVWX__ :
	      ++lu;++lv;++lw;++lx;
              profu=Ubaum->pro[lu];
	      profw=Wbaum->pro[lw];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      infer_star(profx,profv,smx,profu,smu,profw,smw);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUV_X__ :
	      ++lu;++lv;++lx;
              profu=Ubaum->pro[lu];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      infer_three(profx,profv,smx,profu,smu);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case hUVW___ :
	      ++lu;++lv;++lw;
              profu=Ubaum->pro[lu];
	      profw=Wbaum->pro[lw];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Vbaum->pro.push_back(infer_base(profu,smu,profw,smw));
	      break;
	   case hUV____ :
	      ++lu;++lv;
              profu=Ubaum->pro[lu];
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Vbaum->pro.push_back(infer_base(profu,smu));
	      break;
	   case hU_____ :
	      ++lu;
	      Ubaum->nb[u_idx_v].ausrichter.push_back(lv);
	      break;
	   case h_VWXYZ :
	      ++lv;++lw;++lx;++ly;++lz;
	      profw=Wbaum->pro[lw];
	      profy=Ybaum->pro[ly];
	      profz=Zbaum->pro[lz];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_five(profv,profx,smx,profw,smw,profy,smy,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_V_XYZ :
	      ++lv;++lx;++ly;++lz;
	      profy=Ybaum->pro[ly];
	      profz=Zbaum->pro[lz];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_star(profv,profx,smx,profy,smy,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_VWX_Z :
	      ++lv;++lw;++lx;++lz;
	      profw=Wbaum->pro[lw];
	      profz=Zbaum->pro[lz];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_line(profv,profx,smx,profw,smw,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_V_X_Z :
	      ++lv;++lx;++lz;
	      profz=Zbaum->pro[lz];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      infer_three(profv,profx,smx,profz,smz);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_VWXY_ :
	      ++lv;++lw;++lx;++ly;
	      profw=Wbaum->pro[lw];
	      profy=Ybaum->pro[ly];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      infer_line(profv,profx,smx,profw,smw,profy,smy);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_V_XY_ :
	      ++lv;++lx;++ly;
	      profy=Ybaum->pro[ly];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      infer_three(profv,profx,smx,profy,smy);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_VWX__ :
	      ++lv;++lw;++lx;
	      profw=Wbaum->pro[lw];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      infer_three(profx,profv,smx,profw,smw);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_V_X__ :
	      ++lv;++lx;
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      smx.choosepair(profv,profx);
	      Vbaum->pro.push_back(profv);
	      Xbaum->pro.push_back(profx);
	      break;
	   case h_VW___ :
	      ++lv;++lw;
	      profw=Wbaum->pro[lw];
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      Vbaum->pro.push_back(infer_base(profw,smw));
	      break;
	   case h_V____ :
	      ++lv;
	      Vbaum->nb[v_idx_u].ausrichter.push_back(lu);
	      Vbaum->nb[v_idx_w].ausrichter.push_back(lw);
	      Vbaum->nb[v_idx_x].ausrichter.push_back(lx);
	      Vbaum->pro.push_back(sub.choosebase());
	      break;
	   case hW1 :
	   case hW2 :
	   case hW3 :
	   case hW4 :
	   case hW5 :
	      ++lw;
	      Wbaum->nb[w_idx_v].ausrichter.push_back(lv);
	      break;
	   case h___XYZ :
	      ++lx;++ly;++lz;
	      profy=Ybaum->pro[ly];
	      profz=Zbaum->pro[lz];
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      Xbaum->pro.push_back(infer_base(profy,smy,profz,smz));
	      break;
	   case h___X_Z :
	      ++lx;++lz;
	      profz=Zbaum->pro[lz];
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      Xbaum->pro.push_back(infer_base(profz,smz));
	      break;
	   case h___XY_ :
	      ++lx;++ly;
	      profy=Ybaum->pro[ly];
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      Xbaum->pro.push_back(infer_base(profy,smy));
	      break;
	   case h___X__ :
	      ++lx;
	      Xbaum->nb[x_idx_v].ausrichter.push_back(lv);
	      Xbaum->nb[x_idx_y].ausrichter.push_back(ly);
	      Xbaum->nb[x_idx_z].ausrichter.push_back(lz);
	      Xbaum->pro.push_back(sub.choosebase());
	      break;
	   case hY1 :
	   case hY2 :
	      ++ly;
	      Ybaum->nb[y_idx_x].ausrichter.push_back(lx);
	      break;
	   case h_____Z :
	      ++lz;
	      Zbaum->nb[z_idx_x].ausrichter.push_back(lx);
	      break;
	   case hSTART :
              if(hexidx!=0) {
	         cerr << "Late start in function write_centpair\n";
		 exit(EXIT_FAILURE);
	      }
	      break;
	   case hENDE :
              if(hexidx!=allesequs.size()-1) {
	         cerr << "Early end in function write_centpair\n";
		 exit(EXIT_FAILURE);
	      }
	      break;
           default :
	      cerr << "Unknown state in function write_centpair\n";
	      exit(EXIT_FAILURE);
       }
   }
}



/**********  erzeugen von Basen an inneren Knoten  ****************************/

int mult_alignment::infer_base(int profparent, substmodel& smparent,
                               int proflinks, substmodel& smlinks,
                               int profrechts, substmodel& smrechts)
{
   int alf=smparent.retalf();
   int i;
   double logprobsum, probsum2=0.0;
   vector<double> logliks(alf);
   double R=randomunitintervall();

   // Wenn die Aeste sehr kurz sind, wird die gleiche Base genommen
   if(smparent.timeOf()<INFACC*DBL_MIN)
      return profparent;
   if(smlinks.timeOf()<INFACC*DBL_MIN)
      return proflinks;
   if(smrechts.timeOf()<INFACC*DBL_MIN)
      return profrechts;

   for(i=0;i<alf;++i)
       logliks[i]=smparent.logfreqofstate(profparent)+smparent.logProb_ij(profparent,i)
                  +smlinks.logProb_ij(i,proflinks)+smrechts.logProb_ij(i,profrechts);
   logprobsum=addlog(logliks);
   for(i=0;i<alf;++i) {
       probsum2+=exp(logliks[i]-logprobsum);
       if(R<=probsum2)
          return i;
   }
   cerr << "Warning: Reached end of infer_base I\n";
   if(R<1.0/3.0)
      return profparent;
   if(R<2.0/3.0)
      return profrechts;
   return proflinks;
}
 
int mult_alignment::infer_base(int proflinks, substmodel& smlinks,
                               int profrechts, substmodel& smrechts)
{
   int alf=smlinks.retalf();
   int i;
   double logprobsum, probsum2=0.0;
   vector<double> logliks(alf);
   double R=randomunitintervall();
   
   // Wenn die Aeste sehr kurz sind, wird die gleiche Base genommen
   if(smlinks.timeOf()<INFACC*DBL_MIN)
      return proflinks;
   if(smrechts.timeOf()<INFACC*DBL_MIN)
      return profrechts;

   for(i=0;i<alf;++i)
       logliks[i]=smlinks.logfreqofstate(proflinks)+smlinks.logProb_ij(proflinks,i)
                  +smrechts.logProb_ij(i,profrechts);
   logprobsum=addlog(logliks);
   for(i=0;i<alf;++i) {
       probsum2+=exp(logliks[i]-logprobsum);
       if(R<=probsum2)
          return i;
   }
   cerr << "Warning: Reached end of infer_base II\n";
   if(R<0.5)
      return profrechts;
   return proflinks;
}
 
int mult_alignment::infer_base(int profkind, substmodel& smkind)
{
   int alf=smkind.retalf();
   int i;
   double logprobsum, probsum2=0.0;
   vector<double> logliks(alf);
   double R=randomunitintervall();
   
   // Wenn die Aeste sehr kurz sind, wird die gleiche Base genommen
   if(smkind.timeOf()<INFACC*DBL_MIN)
      return profkind;

   for(i=0;i<alf;++i)
       logliks[i]=smkind.logfreqofstate(profkind)+smkind.logProb_ij(profkind,i);
   logprobsum=addlog(logliks);
   for(i=0;i<alf;++i) {
       probsum2+=exp(logliks[i]-logprobsum);
       if(R<=probsum2)
          return i;
   }
   cerr << "Warning: Reached end of infer_base III\n";
   return profkind;
}



void mult_alignment::infer_six(int& vpro, int& xpro, substmodel& svx,
                               int upro, substmodel suv, int wpro, substmodel& svw,
			       int ypro, substmodel sxy, int zpro, substmodel& sxz)
{
   // ermittelt die Profile vpro und xpro fuer den Baum
   // W\   /Y
   //   V-X
   // U/   \Z
   // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   int alf=svx.retalf(); 
   int i,j;
   vector<vector<double> > logliks(alf,vector<double>(alf));
   double logprobsum=-DBL_MAX, probsum2=0.0;
   double R=randomunitintervall();

   vpro=xpro=-1;
   // Wenn die Aeste zu kurz sind, werden die gleichen Basen wie an den Spitzen genommen
   if(svx.timeOf()<INFACC*DBL_MIN) {
      if(R<0.5)
         vpro=infer_base(upro,suv,wpro,svw);
      else
         vpro=infer_base(ypro,sxy,zpro,sxz);
      xpro=vpro;
   } else {
      if(suv.timeOf()<INFACC*DBL_MIN)
         vpro=upro;
      else {
         if(svw.timeOf()<INFACC*DBL_MIN)
            vpro=wpro;
      }
      if(vpro!=-1)
         xpro=infer_base(vpro,svx,ypro,sxy,zpro,sxz);
      else {
         if(sxy.timeOf()<INFACC*DBL_MIN)
            xpro=ypro;
         else {
            if(sxz.timeOf()<INFACC*DBL_MIN)
               xpro=zpro;
         }
         if(xpro!=-1)
            vpro=infer_base(upro,suv,wpro,svw,xpro,svx);
      }
   }

   if(vpro==-1) {
     for(i=0;i<alf;++i) {
       for(j=0;j<alf;++j)
         logliks[i][j]=suv.logfreqofstate(upro)+suv.logProb_ij(upro,i)+svw.logProb_ij(i,wpro)
                      +svx.logProb_ij(i,j)+sxy.logProb_ij(j,ypro)+sxz.logProb_ij(j,zpro);
       logprobsum=addlog(logprobsum,addlog(logliks[i]));
     }
     i=j=0;
     while(vpro==-1 || xpro==-1) {
        if(i<alf) {
           probsum2+=exp(logliks[i][j]-logprobsum);
           if(R<=probsum2) {
              vpro=i;
	      xpro=j;
           }
           ++j;
           if(j==alf) {
              j=0;
	      ++i;
           }
        } else {
           cerr << "Warning: Problems in function infer_six\n";
           if(R<0.25)
              vpro=upro;
	   else {
             if(R<0.5)
                vpro=wpro;
	     else {
               if(R<0.75)
                  vpro=ypro;
	       else {
                  vpro=zpro;
	       }
	     }
	   }
           xpro=vpro;
	}
     }
   }
   if(vpro==-1 || xpro==-1) {
      cerr << "Error while inferring internal bases\n";
      exit(EXIT_FAILURE);
   }
}



void mult_alignment::infer_five(int& vpro, int& xpro, substmodel& svx,
                                int upro, substmodel suv,
			        int ypro, substmodel sxy, int zpro, substmodel& sxz)
{
   // ermittelt die Profile vpro und xpro fuer den Baum
   //      /Y
   //   V-X
   // U/   \Z
   // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   int alf=svx.retalf(); 
   int i,j;
   vector<vector<double> > logliks(alf,vector<double>(alf));
   double logprobsum=-DBL_MAX, probsum2=0.0;
   double R=randomunitintervall();

   vpro=xpro=-1;
   // Wenn die Aeste zu kurz sind, werden die gleichen Basen wie an den Spitzen genommen
   if(svx.timeOf()<INFACC*DBL_MIN) {
     vpro=infer_base(upro,suv,ypro,sxy,zpro,sxz);
     xpro=vpro;
   } else {
      if(suv.timeOf()<INFACC*DBL_MIN)
         vpro=upro;
      if(vpro!=-1)
         xpro=infer_base(vpro,svx,ypro,sxy,zpro,sxz);
      else {
         if(sxy.timeOf()<INFACC*DBL_MIN)
            xpro=ypro;
         else {
            if(sxz.timeOf()<INFACC*DBL_MIN)
               xpro=zpro;
         }
         if(xpro!=-1)
            vpro=infer_base(upro,suv,xpro,svx);
      }
   }

   if(vpro==-1) {
      for(i=0;i<alf;++i) {
        for(j=0;j<alf;++j)
          logliks[i][j]=suv.logfreqofstate(upro)+suv.logProb_ij(upro,i)
                       +svx.logProb_ij(i,j)+sxy.logProb_ij(j,ypro)+sxz.logProb_ij(j,zpro);
        logprobsum=addlog(logprobsum,addlog(logliks[i]));
      }
      i=j=0;
      while(vpro==-1) {
         if(i<alf) {
            probsum2+=exp(logliks[i][j]-logprobsum);
            if(R<=probsum2) {
               vpro=i;
	       xpro=j;
            }
            ++j;
            if(j==alf) {
               j=0;
	       ++i;
            }
         } else {
            cerr << "Warning: Problems in function infer_five\n";
            if(R<1.0/3.0)
              vpro=upro;
	    else {
              if(R<2.0/3.0)
                vpro=ypro;
	      else {
                  vpro=zpro;
	      }
	    }
	    xpro=vpro;
         }
      }
   }
   if(vpro==-1 || xpro==-1) {
      cerr << "Error while inferring internal bases\n";
      exit(EXIT_FAILURE);
   }
}

void mult_alignment::infer_line(int& vpro, int& xpro, substmodel& svx,
                                int upro, substmodel suv, int zpro, substmodel& sxz)
{
   // ermittelt die Profile vpro und xpro fuer den Baum
   //      
   //   V-X
   // U/   \Z
   // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   int alf=svx.retalf(); 
   int i,j;
   vector<vector<double> > logliks(alf,vector<double>(alf));
   double logprobsum=-DBL_MAX, probsum2=0.0;
   double R=randomunitintervall();

   vpro=xpro=-1;
   // Wenn die Aeste zu kurz sind, werden die gleichen Basen wie an den Spitzen genommen
   if(svx.timeOf()<INFACC*DBL_MIN) {
     vpro=infer_base(upro,suv,zpro,sxz);
     xpro=vpro;
   } else {
      if(suv.timeOf()<INFACC*DBL_MIN)
         vpro=upro;
      if(vpro!=-1)
         xpro=infer_base(vpro,svx,zpro,sxz);
      else {
         if(sxz.timeOf()<INFACC*DBL_MIN)
            xpro=zpro;
         if(xpro!=-1)
            vpro=infer_base(upro,suv,xpro,svx);
      }
   }

   if(vpro==-1) {
      for(i=0;i<alf;++i) {
        for(j=0;j<alf;++j)
         logliks[i][j]=suv.logfreqofstate(upro)+suv.logProb_ij(upro,i)
                       +svx.logProb_ij(i,j)+sxz.logProb_ij(j,zpro);
        logprobsum=addlog(logprobsum,addlog(logliks[i]));
      }
      i=j=0;
      while(vpro==-1) {
         if(i<alf) {
            probsum2+=exp(logliks[i][j]-logprobsum);
            if(R<=probsum2) {
               vpro=i;
	       xpro=j;
            }
            ++j;
            if(j==alf) {
               j=0;
	       ++i;
            }
         } else {
            cerr << "Warning: Problems in function infer_line\n";
            if(R<0.5)
              xpro=upro;
	    else
              xpro=zpro;
	    vpro=xpro;
	 }
      }
   }
   if(vpro==-1 || xpro==-1) {
      cerr << "Error while inferring internal bases\n";
      exit(EXIT_FAILURE);
   }
}

void mult_alignment::infer_star(int& vpro, int& xpro, substmodel& svx,
			        int ypro, substmodel sxy, int zpro, substmodel& sxz)
{
   // ermittelt die Profile vpro und xpro fuer den Baum
   //      /Y
   //   V-X
   //      \Z
   // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   int alf=svx.retalf(); 
   int i,j;
   vector<vector<double> > logliks(alf,vector<double>(alf));
   double logprobsum=-DBL_MAX, probsum2=0.0;
   double R=randomunitintervall();

   vpro=xpro=-1;
   // Wenn die Aeste zu kurz sind, werden die gleichen Basen wie an den Spitzen genommen
   if(sxz.timeOf()<INFACC*DBL_MIN)
      xpro=zpro;
   else {
      if(sxy.timeOf()<INFACC*DBL_MIN)
         xpro=ypro;
      else {
         if(svx.timeOf()<INFACC*DBL_MIN)
	    xpro=infer_base(ypro,sxy,zpro,sxz);
      }
   }
   if(xpro!=-1)
      vpro=infer_base(xpro,svx);

   if(vpro==-1) {
      for(i=0;i<alf;++i) {
        for(j=0;j<alf;++j)
          logliks[i][j]=svx.logfreqofstate(i)+svx.logProb_ij(i,j)
                       +sxy.logProb_ij(j,ypro)+sxz.logProb_ij(j,zpro);
        logprobsum=addlog(logprobsum,addlog(logliks[i]));
      }
      i=j=0;
      while(vpro==-1) {
         if(i<alf) {
            probsum2+=exp(logliks[i][j]-logprobsum);
            if(R<=probsum2) {
               vpro=i;
	       xpro=j;
            }
            ++j;
            if(j==alf) {
               j=0;
	       ++i;
            }
         } else {
            cerr << "Warning: Problems in function infer_star\n";
            if(R<0.5)
              xpro=ypro;
            else
              xpro=zpro;
	    vpro=xpro;
	 }
      }
   }
   if(vpro==-1 || xpro==-1) {
      cerr << "Error while inferring internal bases\n";
      exit(EXIT_FAILURE);
   }
}


void mult_alignment::infer_three(int& vpro, int& xpro, substmodel& svx,
                                 int zpro, substmodel& sxz)
{
   // ermittelt die Profile vpro und xpro fuer den Baum
   //      
   //   V-X
   //      \Z
   // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   int alf=svx.retalf(); 
   int i,j;
   vector<vector<double> > logliks(alf,vector<double>(alf));
   double logprobsum=-DBL_MAX, probsum2=0.0;
   double R=randomunitintervall();

   vpro=xpro=-1;
   // Wenn die Aeste zu kurz sind, werden die gleichen Basen wie an den Spitzen genommen
   if(sxz.timeOf()<INFACC*DBL_MIN)
      xpro=zpro;
   else {
      if(svx.timeOf()<INFACC*DBL_MIN)
         xpro=infer_base(zpro,sxz);
   }
   if(xpro!=-1)
      vpro=infer_base(xpro,svx);

   if(vpro==-1) {
      for(i=0;i<alf;++i) {
        for(j=0;j<alf;++j)
          logliks[i][j]=svx.logfreqofstate(i)+svx.logProb_ij(i,j)+sxz.logProb_ij(j,zpro);
        logprobsum=addlog(logprobsum,addlog(logliks[i]));
      }
      i=j=0;
      while(vpro==-1) {
         if(i<alf) {
            probsum2+=exp(logliks[i][j]-logprobsum);
            if(R<=probsum2) {
               vpro=i;
	       xpro=j;
            }
            ++j;
            if(j==alf) {
               j=0;
	       ++i;
            }
         } else {
            cerr << "Warning: Problems in function infer_three\n";
	    vpro=zpro; xpro=vpro;
         }
      }
   }
   if(vpro==-1 || xpro==-1) {
      cerr << "Error while inferring internal bases\n";
      exit(EXIT_FAILURE);
   }
}



/**********  Verschiedenes      ************************/
 
void mult_alignment::ausaltmachneu()
{
   if(tnew!=NULL) {
      if(!tnew->isRoot()) {
          tnew->reroot(tnew);
	  cerr << "WARNING: tnew's root has changed\n";
      }
      delete(tnew);
   }
   
   tnew = new baum(*told);
}


void mult_alignment::ausneumachalt()
{
   if(told!=NULL) {
      if(!told->isRoot()) {
          told->reroot(told);
	  cerr << "WARNING: told's root has changed\n";
      }
      delete(told);
   }
   
   told = new baum(*tnew);
}

void mult_alignment::familie(vector<unsigned int>& stern, baum *pt)
{
         if(pt==NULL) {
	     cerr << "The tree in function familie is a NULL pointer!\n";
	     exit(EXIT_FAILURE);
	 }
         stern.clear();
         unsigned int p=4,l=4,r=4;
	 if(pt->nb.size()!=3) {
	     cerr << "An internal node turned out not to be internal!\n"
	          << "nb.size()="<<pt->nb.size()<<endl;
	     exit(EXIT_FAILURE);
	 }
         for(unsigned int j=0;j<3;++j) {
             if(pt->nb[j].toroot) {
               if(p!=4) {
	          cerr << "More than one neighbour lies towards the root!\n";
		  exit(EXIT_FAILURE);
	       }
	       p=j;
             } else {
               if(l==4)
	          l=j;
	       else
	          if(r==4)
	             r=j;
		  else {
		     cerr << "Error while looking for grand parent and children!\n";
		     exit(EXIT_FAILURE);
		  }
             }
         }
	 stern.push_back(p);
	 stern.push_back(l);
	 stern.push_back(r);
}


void mult_alignment::familie(vector<unsigned int>& stern, baum *p, baum *q)
{
  // ist q nicht der Vater wird's das rechte Kind
  familie(stern,p);
  if(p->whichone(q)==(int)stern[1]) {
     unsigned int ablage=stern[2];
     stern[2]=stern[1];
     stern[1]=ablage;
  }
}


void mult_alignment::keeptrack(vector<baum *>& nodes, vector<unsigned int>& nbidx,
                               baum *p, int ownch, baum *q, int nbch)
// Die Kante zwischen q und seinem nbch-ten Kind
// wird jetzt durch p und dessen ownch-ten Kind beschrieben
{
    for(unsigned int i=0;i<nodes.size();++i) {
        if(nodes[i]==q && (int)nbidx[i]==nbch) {
           nodes[i]=p;
	   nbidx[i]=ownch;
	} else {
           if(nodes[i]==p && (int)nbidx[i]==ownch) {
	      nodes[i]=q;
	      nbidx[i]=nbch;
	   }
	}
    }
}
 



/*************   Funktionen fuer das Viereralignment  ***********/

void mult_alignment::quart_to_tri(vector<quart_zus>& quart,vector<triplet_zus>& tri)
{
   // ermittelt das Alignment der drei aeusseren Sequenzen
   unsigned int i;
   if(!tri.empty())
      tri.clear();
   for(i=0;i<quart.size();i++)
      switch(quart[i]) {
          case qSTART :
	     if(i!=0) {
	        cerr << "Late start of vector quart in quart_to_tri\n";
		exit(EXIT_FAILURE);
	     }
	     tri.push_back(tSTART);
	     break;
          case qWXYZ :
	     tri.push_back(tXYZ);
	     break;
          case qWX_Z :
	     tri.push_back(tX_Z);
	     break;
          case qWXY_ :
	     tri.push_back(tXY_);
	     break;
          case qWX__ :
          case qW___ :
	     tri.push_back(tX__);
	     break;
          case q_XY_ :
          case q__Y_1 :
          case q__Y_2 :
          case q__Y_3 :
          case q__Y_4 :
	     tri.push_back(t_Y_1);
	     break;
          case q_X_Z :
          case q___Z1 :
          case q___Z2 :
	     tri.push_back(t__Z);
	     break;
          case q_XYZ :
	     tri.push_back(t_YZ);
	     break;
          case q_X__ :
	     break;
          case qENDE :
	     if(i!=quart.size()-1) {
	        cerr << "Premature end of vector quart in quart_to_tri\n";
		exit(EXIT_FAILURE);
	     }
	     tri.push_back(tENDE);
	     break;
	  default :
	     cerr << "Strange state in function quart_to_tri\n";
	     exit(EXIT_FAILURE);
      }
      
   if(tri.back()!=tENDE) {
      cerr << "Error in function quart_to_tri\n";
      exit(EXIT_FAILURE);
   }
}

void mult_alignment::quartali(vector<quart_zus>& quart, baum *b, vector<unsigned int>& Kinder)
// ermittelt das Alignment am Mercedesstern um Knoten b
{
   vector<vector<int> > WYZ(3,vector<int>(2,-1)), altWYZ(3); // WYZ[i][0] ist der Index der zentralen Sequenz
                                                          // im Alignment mit ihrem i-ten Kind; WYZ[i][1] 
							  // der des Kindes. altWYZ speichert die alten Werte.
   vector<baum *> neighbours(3); // Zeiger auf die Nachbarn
   vector<unsigned int> bidx(3); // Index des zentralen Knotens bei den Nachbarn
   vector<zustand> paar_zus(3);  // Zustaende der paarweisen Alignments
   bool bdummy=TRUE;
   vector<bool> aktiv(3,bdummy);   // Welche paarweisen Alignments haben Einfluss auf das Dreieralignment
   int i,j,minx=0,idxminx;       // minx ist der kleinste Laufindex der Sequenz am zentralen Knoten
                                 // idxminx der Index des dazugehoerigen Alignments

   if(b==NULL || Kinder.size()!=3) {  // nur zur Sicherheit
      cerr << "Unreasonable arguments to quartali!\n";
      exit(EXIT_FAILURE);
   }
   
   for(i=0;i<3;++i) { // Wo sind die Nachbarn
      neighbours[i]=b->nb[Kinder[i]].nachbar;
      if(neighbours[i]==NULL) {
         cerr << "Encountered NULL pointer\n";
	 exit(EXIT_FAILURE);
      }
      bidx[i]=neighbours[i]->whichone(b);
   }
   if(!quart.empty())  // Viereralignment leeren
      quart.clear();
   do { // bis ans Ende des Viereralignments
     for(i=0;i<3;i++) // nur in Alignments, die zum aktuellen Zustand beigetragen haben,
       if(aktiv[i]) {  //  wird nach dem naechsten Zustand gefragt
         altWYZ[i]=WYZ[i];
         paar_zus[i]=iter_zust(b->nb[Kinder[i]].ausrichter,neighbours[i]->nb[bidx[i]].ausrichter,WYZ[i]);
       }
     // finde groessten Index mit kleinstem Laufindex von X
     idxminx=0;
     minx=altWYZ[0][0];
     for(j=1;j<3;++j)
         if(altWYZ[j][0]<=minx && paar_zus[j]!=ENDE) { 
	    minx=altWYZ[j][0];
	    idxminx=j;
         }

     switch(paar_zus[idxminx]) {
       case START :  // alle paarweisen Alignments muessen im Zustand START sein
         for(j=0;j<3;++j)
	   if(paar_zus[j]!=START) {
	      cerr << "Error: pairwise alignments are asynchronous I\n";
	      exit(EXIT_FAILURE);
	   }
	 quart.push_back(qSTART);
         break;
       case ENDE :   // alle muessen im Zustand ENDE angekommen sein
         for(j=0;j<3;++j) {
	   if(paar_zus[j]!=ENDE) {
	      cerr << "Error: pairwise alignments are asynchronous II\n";
	      exit(EXIT_FAILURE);
	   }
	 }  
	 quart.push_back(qENDE);
         break;
       case PAIR :   // alle muessen entweder im Zustand PAIR oder DELETION sein
         for(j=0;j<3;++j) {
	   if(paar_zus[j]!=PAIR && paar_zus[j]!=DELETION) {
	      cerr << "Error: pairwise alignments are asynchronous III\n";
	      exit(EXIT_FAILURE);
	   }
	   aktiv[j]=TRUE;
	 }
	 quart.push_back(which_quartet_zus(paar_zus));
         break;
       case INSERTION :  // alles moeglich bis auf START
         if(idxminx==2 && altWYZ[1][0]==minx && paar_zus[1]==INSERTION) // legt die Reihenfolge der Insertionen fest (links vor rechts)
	     idxminx=1;
         for(j=0;j<3;j++)
	     if(j==idxminx)
	        aktiv[j]=TRUE;
	     else
                aktiv[j]=FALSE;
	 switch (idxminx) {
	    case 0 :
	       quart.push_back(qW___);
	       break;
	    case 1 :
	       switch(quart.back()) {
	          case qSTART :
		  case q_XYZ :
		  case q_X_Z :
		  case q__Y_1 :
		    quart.push_back(q__Y_1);
		    break;
		  case q_XY_ :
		  case q_X__ :
		  case q__Y_2 :
		    quart.push_back(q__Y_2);
		    break;
		  case qWXYZ :
		  case qWX_Z :
		  case q__Y_3 :
		    quart.push_back(q__Y_3);
		    break;
		  case qWXY_ :
		  case qWX__ :
		  case q__Y_4 :
		    quart.push_back(q__Y_4);
		    break;
	          default :
		    cerr << "Nonsense in function quartali\n";
		    exit(EXIT_FAILURE);
	       }
	       break;
	    case 2 :
	       switch(quart.back()) {
	          case qSTART :
		  case q_XYZ :
		  case q_X_Z :
		  case q__Y_1 :
		  case q_XY_ :
		  case q_X__ :
		  case q__Y_2 :
		  case q___Z1 :
		    quart.push_back(q___Z1);
		    break;
		  case qWXYZ :
		  case qWX_Z :
		  case q__Y_3 :
		  case qWXY_ :
		  case qWX__ :
		  case q__Y_4 :
		  case q___Z2 :
		    quart.push_back(q___Z2);
		    break;
	          default :
		    cerr << "Nonsense in function quartali\n";
		    exit(EXIT_FAILURE);
	       }
	       break;
	    default :
	       cerr << "Nonsense\n";
	       exit(EXIT_FAILURE);
	 }
         break;
       case DELETION : // alle muessen entweder im Zustand PAIR oder DELETION sein
         for(j=0;j<3;++j) {
	   if(paar_zus[j]!=PAIR && paar_zus[j]!=DELETION) {
	      cerr << "Error: pairwise alignments are asynchronous IV\n";
	      exit(EXIT_FAILURE);
	   }
	   aktiv[j]=TRUE;
	 }
	 quart.push_back(which_quartet_zus(paar_zus));
         break;
       default :
         cerr << "Unreasonable state in function quartali!\n";
	 exit(EXIT_FAILURE);
     }

   }while(quart.back()!=qENDE);

}


quart_zus mult_alignment::which_quartet_zus(vector<zustand>& paar_zus)
{
   quart_zus qu=qNOSTATE;
   
   if(paar_zus.size()!=3) {
      cerr << "A strange argument was passed to which_quartet_zus!\n";
      exit(EXIT_FAILURE);
   }
   
   switch(paar_zus[0]) {
      case PAIR :
        switch(paar_zus[1]) {
              case PAIR :
	        switch(paar_zus[2]) {
		   case PAIR :
		      qu = qWXYZ;
		      break;
		   case DELETION :
		      qu = qWXY_;
		      break;
                   default :
                      cerr << "Unreasonable state in function which_quartet_zus\n";
	              exit(EXIT_FAILURE);
		}
                break;
              case DELETION :
	        switch(paar_zus[2]) {
		   case PAIR :
		      qu = qWX_Z;
		      break;
		   case DELETION :
		      qu = qWX__;
		      break;
                   default :
                      cerr << "Unreasonable state in function which_quartet_zus\n";
	              exit(EXIT_FAILURE);
		}
                break;
              default :
                cerr << "Unreasonable state in function which_quartet_zus\n";
	        exit(EXIT_FAILURE);
	}
        break;
      case DELETION :
        switch(paar_zus[1]) {
              case PAIR :
	        switch(paar_zus[2]) {
		   case PAIR :
		      qu = q_XYZ;
		      break;
		   case DELETION :
		      qu = q_XY_;
		      break;
                   default :
                      cerr << "Unreasonable state in function which_quartet_zus\n";
	              exit(EXIT_FAILURE);
		}
                break;
              case DELETION :
	        switch(paar_zus[2]) {
		   case PAIR :
		      qu = q_X_Z;
		      break;
		   case DELETION :
		      qu = q_X__;
		      break;
                   default :
                      cerr << "Unreasonable state in function which_quartet_zus\n";
	              exit(EXIT_FAILURE);
		}
                break;
              default :
                cerr << "Unreasonable state in function which_quartet_zus\n";
	        exit(EXIT_FAILURE);
	}
        break;
      default :
        cerr << "Unreasonable state in function which_quartet_zus\n";
	exit(EXIT_FAILURE);
   }
   
   return qu;
}




/*************   Funktionen fuer das Sechseralignment ***********/

void mult_alignment::hexali(baum *p, baum *q, int pidx, int qidx)
{
     vector<unsigned int> pfam, qfam;
     vector<quart_zus> pquart, qquart;

     familie(pfam,p,q); // ist q nicht der Vater von p wird's das rechte Kind
     familie(qfam,q,p); // ist p nicht der Vater von q wird's das rechte Kind
     
     quartali(pquart,p,pfam);
     quartali(qquart,q,qfam);
     mercedes_stern m1(p,pquart,pfam,pidx);
     mercedes_stern m2(q,qquart,qfam,qidx);

     // neue zentrale Profile
     central_pair(m1,m2);
}

swapflag mult_alignment::which_pair_swaps(mercedes_stern& m1, mercedes_stern& m2,
                                          mercedes_stern*& anc, mercedes_stern*& rec)
{
    swapflag swp;
    
    if(m1.tree->parent()==m2.tree) {
       anc=&m2;
       rec=&m1;
    } else 
      if(m2.tree->parent()==m1.tree) {
       anc=&m1;
       rec=&m2;
    }
    
    unsigned int widx, yidx, zidx;
    if(anc->tree->whichone(rec->tree)==(int)anc->ordered_children[1]) {   
        widx=anc->ordered_children[2]; 
        cerr << "VORSICHT, rec sollte nicht das linke Kind von anc sein!!!!!\n";
    }
    else {
       if(anc->tree->whichone(rec->tree)==(int)anc->ordered_children[2])
          widx=anc->ordered_children[1];
       else {
          cerr << "Error I in function which_pair_swaps\n";
	  exit(EXIT_FAILURE);
       }
    }
    if(rec->tree->whichone(anc->tree)!=(int)rec->ordered_children[0]) {
       cerr << "Error II in function which_pair_swaps\n";
       exit(EXIT_FAILURE);
    }
    yidx=rec->ordered_children[1];
    zidx=rec->ordered_children[2];
    
    if(anc->verstossen==(int)anc->ordered_children[0]) {
          if(rec->verstossen==(int)yidx)
	     swp=WggY;
	  else {
	     if(rec->verstossen==(int)zidx)
	        swp=WggZ;
	     else {
	        cerr << "Error III in function which_pair_swaps\n";
		exit(EXIT_FAILURE);
	     }
	  }
    } else {
       if(anc->verstossen==(int)widx) {
          if(rec->verstossen==(int)yidx)
	     swp=XggY;
	  else {
	     if(rec->verstossen==(int)zidx)
	        swp=XggZ;
	     else {
	        cerr << "Error IV in function which_pair_swaps\n";
		exit(EXIT_FAILURE);
	     }
	  }
       } else {
           cerr << "Error V in function which_pair_swaps\n";
	   exit(EXIT_FAILURE);
       }
    }

    return swp;
}

void mult_alignment::fillH(mercedes_stern& anc, mercedes_stern& rec, swapflag swp)
{
   if(rec.tree->parent()!=anc.tree) {
      cerr << "anc is not the parent of rec in function fillH\n";
      exit(EXIT_FAILURE);
   }

   double tuv, tvw, tvx, txy, txz, ablage;

    if(anc.tree->whichone(rec.tree)==(int)anc.ordered_children[1]) {
       tvw=anc.tree->nb[anc.ordered_children[2]].length;
       tvx=anc.tree->nb[anc.ordered_children[1]].length;
    } else {
       if(anc.tree->whichone(rec.tree)==(int)anc.ordered_children[2]) {
          tvw=anc.tree->nb[anc.ordered_children[1]].length;
          tvx=anc.tree->nb[anc.ordered_children[2]].length;
       } else {
          cerr << "Error I in function fillH\n";
	  exit(EXIT_FAILURE);
       }
    }
    if(rec.tree->whichone(anc.tree)!=(int)rec.ordered_children[0]) {
       cerr << "Error II in function fillH\n";
       exit(EXIT_FAILURE);
    }
    tuv=anc.tree->nb[anc.ordered_children[0]].length;
    txy=rec.tree->nb[rec.ordered_children[1]].length;
    txz=rec.tree->nb[rec.ordered_children[2]].length;

   switch(swp) {// W bleibt immer W!!!
     case WggZ :
     case XggY :
        ablage=txy;
        txy=tvw;
	tvw=ablage;
        break;
     case WggY :
     case XggZ :
        ablage=txz;
        txz=tvw;
	tvw=ablage;
        break;
     default :
        cerr << "Impossible value of swp\n";
	exit(EXIT_FAILURE);
   }
   id.fillH(tuv,tvw,tvx,txy,txz);
}



void mult_alignment::zweimalvier(vector<quart_zus>& quart, mercedes_stern& m1, mercedes_stern& m2, swapflag swp)
{
   // Setzt die Alignments zweier benachbarter Mercedessterne zu einem
   // Alignment der vier aeusseren Sequenzen zusammen und beruecksichtigt
   // dabei, welche Aeste spaeter miteinander vertauscht werden.
   
   quart_zus qz;
   unsigned int idummy=0;
   vector<unsigned int> idx(2,idummy);
   
   if(!quart.empty())
      quart.clear();
   
   do {
      qz=swap_quart_zus(hex_to_quart(two_quart_to_hex(m1,m2,idx)),swp);
      if(qz!=qNOSTATE)
         quart.push_back(qz);
   }while(quart.back()!=qENDE && idx[0]<m1.ali.size() && idx[1]<m2.ali.size());
   
   if(idx[0]!=m1.ali.size() || idx[1]!=m2.ali.size() || quart.back()!=qENDE) {
      cerr << "Error in function zweimalvier\n";
      exit(EXIT_FAILURE);
   }
}


hex_zus mult_alignment::two_quart_to_hex(mercedes_stern& m1, mercedes_stern& m2, vector<unsigned int>& idx)
{
   // Zentralknoten von m1 muss der Vater von dem von m2 sein!
   // Nimmt keine Ruecksicht auf die verschiedenen qY,qZ, hW und hY Zustaende!
   // Darf nicht fuer die Berechnung von Likelihoods verwendet werden!
   // aus X in m1.ali folgt W in m2.ali (und umgekehrt)
   // aus Z in m1.ali folgt X in m2.ali (und umgekehrt)
   
   hex_zus hx=hNOSTATE;
   
   if(idx.size()!=2) {
      cerr << "Inadequate index vector in function two_quart_to_hex!\n";
      exit(EXIT_FAILURE);
   }
   
   switch(m1.ali[idx[0]]) {
      case qWXYZ :
        switch(m2.ali[idx[1]]) {
           case qWXYZ :
	     hx=hUVWXYZ;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWXY_ :
	     hx=hUVWXY_;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWX_Z :
	     hx=hUVWX_Z;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWX__ :
	     hx=hUVWX__;
	     ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case qWXY_ :
        switch(m2.ali[idx[1]]) {
           case qW___ :
             hx=hUVW___;
             ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case qWX_Z :
        switch(m2.ali[idx[1]]) {
           case qWXYZ :
	     hx=hUV_XYZ;
             ++idx[0];
             ++idx[1];
             break;
           case qWXY_ :
	     hx=hUV_XY_;
             ++idx[0];
             ++idx[1];
             break;
           case qWX_Z :
	     hx=hUV_X_Z;
             ++idx[0];
             ++idx[1];
             break;
           case qWX__ :
	     hx=hUV_X__;
             ++idx[0];
             ++idx[1];
             break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case q_XYZ :
        switch(m2.ali[idx[1]]) {
           case qWXYZ :
	     hx=h_VWXYZ;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWXY_ :
	     hx=h_VWXY_;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWX_Z :
	     hx=h_VWX_Z;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWX__ :
	     hx=h_VWX__;
	     ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case qWX__ :
        switch(m2.ali[idx[1]]) {
           case qW___ :
             hx=hUV____;
             ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case q_XY_ :
        switch(m2.ali[idx[1]]) {
           case qW___ :
             hx=h_VW___;
             ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case q_X_Z :
        switch(m2.ali[idx[1]]) {
           case qWXYZ :
	     hx=h_V_XYZ;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWXY_ :
	     hx=h_V_XY_;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWX_Z :
	     hx=h_V_X_Z;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case qWX__ :
	     hx=h_V_X__;
	     ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case qW___ :
        switch(m2.ali[idx[1]]) {
           case qWXYZ :
	   case qWXY_ :
	   case qWX_Z :
           case q_XYZ :
	   case qWX__ :
           case q_XY_ :
           case q_X_Z :
	   case qW___ :
           case q_X__ :
           case qSTART :
	   case qENDE :
             hx=hU_____;
             ++idx[0];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case q_X__ :
        switch(m2.ali[idx[1]]) {
           case qW___ :
             hx=h_V____;
             ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case q__Y_1 :
      case q__Y_2 :
      case q__Y_3 :
      case q__Y_4 :   // links vor rechts
        switch(m2.ali[idx[1]]) {
           case qWXYZ :
	   case qWXY_ :
	   case qWX_Z :
           case q_XYZ :
	   case qWX__ :
           case q_XY_ :
           case q_X_Z :
	   case qW___ :
           case q_X__ :
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
      	   case q___Z1 :
      	   case q___Z2 :
           case qSTART :
	   case qENDE :
             hx=hW1;
             ++idx[0];
	     break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case q___Z1 :
      case q___Z2 :
        switch(m2.ali[idx[1]]) {
           case q_XYZ :
	     hx=h___XYZ;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case q_XY_ :
	     hx=h___XY_;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case q_X_Z :
	     hx=h___X_Z;
	     ++idx[0];
	     ++idx[1];
	     break;
	   case q_X__ :
	     hx=h___X__;
	     ++idx[0];
	     ++idx[1];
	     break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case qSTART :
        switch(m2.ali[idx[1]]) {
           case qSTART :
             hx=hSTART;
             ++idx[0];
             ++idx[1];
             break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      case qENDE :
        switch(m2.ali[idx[1]]) {
           case qENDE :
             hx=hENDE;
             ++idx[0];
             ++idx[1];
             break;
      	   case q__Y_1 :
      	   case q__Y_2 :
      	   case q__Y_3 :
      	   case q__Y_4 :
	     hx=hY1;
	     ++idx[1];
             break;
      	   case q___Z1 :
      	   case q___Z2 :
	     hx=h_____Z;
	     ++idx[1];
             break;
	   default :
	      cerr << "Unreasonable state in m2.ali\n";
	      exit(EXIT_FAILURE);
	}
        break;
      default :
         cerr << "Unreasonable state in m1.ali\n";
	 exit(EXIT_FAILURE);
   }
   
   return hx;
}


quart_zus mult_alignment::hex_to_quart(hex_zus hex)
{
   // gibt den Alignment zustand der vier aeusseren Knoten zurueck
   // Vertraegt sich nicht mit der Berechnung der Likelihood, da nicht
   // zwischen den verschiedenen qY und qZ Zustaenden unterschieden wird!
   quart_zus quart;
   
   switch(hex) {
      case hSTART :
        quart=qSTART;
	break;
      case hENDE :
        quart=qENDE;
        break;
      case h_V_X__ :
      case h_V____ :
      case h___X__ :
        quart=qNOSTATE;
	break;
      case hUVWXYZ :
        quart=qWXYZ;
	break;
      case hUV_XYZ :
        quart=qW_YZ;
	break;
      case hUVWX_Z :
        quart=qWX_Z;
	break;
      case hUV_X_Z :
        quart=qW__Z;
	break;
      case hUVWXY_ :
        quart=qWXY_;
	break;
      case hUV_XY_ :
        quart=qW_Y_;
	break;
      case hUVWX__ :
      case hUVW___ :
        quart=qWX__;
	break;
      case hUV_X__ :
      case hUV____ :
      case hU_____ :
        quart=qW___;
	break;
      case h_VWXYZ :
        quart=q_XYZ;
	break;
      case h_V_XYZ :
      case h___XYZ :
        quart=q__YZ;
	break;
      case h_VWX_Z :
        quart=q_X_Z;
	break;
      case h_V_X_Z :
      case h___X_Z :
      case h_____Z :
        quart=q___Z1;
	break;
      case h_VWXY_ :
        quart=q_XY_;
	break;
      case h_VWX__ :
      case h_VW___ :
      case hW1 :
      case hW2 :
      case hW3 :
      case hW4 :
      case hW5 :
        quart=q_X__;
        break;
      case h_V_XY_ :
      case h___XY_ :
      case hY1 :
      case hY2 :
        quart=q__Y_1;
        break;
      default :
        cerr << "Strange hextett state in function hex_to_quart\n";
	exit(EXIT_FAILURE);
   }
   
   return quart;
}

quart_zus mult_alignment::swap_quart_zus(quart_zus alt,swapflag swp)
{
   // Gibt neuen Zustand eine Viereralignments zurueck,
   // wenn die Topologie geaendert wurde (entspricht hier einer Umbenennung der Blaetter).
   // Die Wurzel (d.h. W) behaelt dabei immer ihren Namen. Deshalb fallen die Faelle
   // WggZ und XggY bzw. WggY und XggZ zusammen.
   // Vertraegt sich nicht mit der Berechnung der Likelihood, da nicht
   // zwischen den verschiedenen qY und qZ Zustaenden unterschieden wird!

   quart_zus neu;
   
   switch(alt) {
      case qSTART :
      case qENDE :
      case qNOSTATE :
      case qWXYZ :
      case q_XYZ :
      case qW___ :
        neu=alt;
	break;
      case qWXY_ :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=alt;
	     break;
	   case WggY :
	   case XggZ :
	     neu=qW_YZ;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case qWX_Z :
        switch (swp) {
	   case WggY :
	   case XggZ :
	     neu=alt;
	     break;
	   case WggZ :
	   case XggY :
	     neu=qW_YZ;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case qWX__ :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=qW_Y_;
	     break;
	   case XggZ :
	   case WggY :
	     neu=qW__Z;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case q_XY_ :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=alt;
	     break;
	   case WggY :
	   case XggZ :
	     neu=q__YZ;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case q_X_Z :
        switch (swp) {
	   case WggY :
	   case XggZ :
	     neu=alt;
	     break;
	   case WggZ :
	   case XggY :
	     neu=q__YZ;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case q_X__ :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=q__Y_1;
	     break;
	   case WggY :
	   case XggZ :
	     neu=q___Z1;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case q__Y_1 :
      case q__Y_2 :
      case q__Y_3 :
      case q__Y_4 :
        switch (swp) {
	   case WggY :
	   case XggZ :
	     neu=alt;
	     break;
	   case WggZ :
	   case XggY :
	     neu=q_X__;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case q___Z1 :
      case q___Z2 :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=alt;
	     break;
	   case XggZ :
	   case WggY :
	     neu=q_X__;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case qW_YZ :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=qWX_Z;
	     break;
	   case WggY :
	   case XggZ :
	     neu=qWXY_;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case qW_Y_ :
        switch (swp) {
	   case WggY :
	   case XggZ :
	     neu=alt;
	     break;
	   case WggZ :
	   case XggY :
	     neu=qWX__;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case qW__Z :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=alt;
	     break;
	   case WggY :
	   case XggZ :
	     neu=qWX__;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      case q__YZ :
        switch (swp) {
	   case WggZ :
	   case XggY :
	     neu=q_X_Z;
	     break;
	   case WggY :
	   case XggZ :
	     neu=q_XY_;
	     break;
	   default :
	     cerr << "swapflag="<<(int)swp<<endl;
	     exit(EXIT_FAILURE);
	}
	break;
      default :
        cerr << "Strange state in function swap_quart_zus\n";
	exit(EXIT_FAILURE);
   }
   
   return neu;
}




/*************************************************************/
/****************** class mercedes_stern   *******************/
/*************************************************************/

mercedes_stern::mercedes_stern()
{
   cerr << "Leerer Mercedesstern unsinnig\n";
   exit(EXIT_FAILURE);
}


mercedes_stern::mercedes_stern(baum *p, vector<quart_zus>& vieral, vector<unsigned int>& kd, int childtochange)
{
   if(p==NULL || p->nb.size()!=3) {
      cerr << "Unreasonable tree pointer in constructor of mercedes_stern\n";
      exit(EXIT_FAILURE);
   }
   tree=p;
   ali=vieral;
   if(kd.size()!=3) {
      cerr << "Unreasonable vector kd in constructor of mercedes_stern\n";
      exit(EXIT_FAILURE);   
   }
   ordered_children=kd;
   if(childtochange<0 || childtochange>2) {
      cerr << "Unreasonable value of childtochange in constructor of mercedes_stern\n";
      exit(EXIT_FAILURE);   
   }
   verstossen=childtochange;
}



/*************************************************************/
/****************** class mdataset         *******************/
/*************************************************************/

mdataset::mdataset()
{
}

mdataset::mdataset(baum *b, alphsize alf)
{  
   vector<int> ali_to_node;
   vector<int> node_to_ali;
   typ=alf;
   push_mult(b,ali_to_node,node_to_ali,Identif,Seqs);
}

void mdataset::operator=(mdataset md2)
{
   Identif=md2.Identif;
   Seqs=md2.Seqs;
   typ=md2.typ;
}

void mdataset::push_mult(baum *b, vector<int>& ali_to_node, vector<int>& node_to_ali,
                         vector<vector<char> >& idmat, vector<vector<char> >& seqmat)
{
   vector<vector<vector<int> > > kaus(b->nb.size(),vector<vector<int> >(2)); // Ausrichter der Kinderalignments
   vector<vector<vector<char> > > Namen(b->nb.size());
   vector<vector<vector<char> > > Sequs(b->nb.size());
   unsigned int i, j;
   vector<vector<int> > atn(b->nb.size(),vector<int>(2,-1)),atnalt;
   vector<zustand> zus(b->nb.size());  // enthaelt die Zustaende der paarweisen Alignments
   vector<char> column;  // eine Spalte des zu bastelnden Alignments
   
   if(b==NULL) {
      cerr << "Error in function push_mult: tree points to NULL\n";
      exit(EXIT_FAILURE);
   }
  
   for(i=0;i<b->nb.size();++i) {
     if(!b->nb[i].toroot) {
        push_mult(b->nb[i].nachbar,kaus[i][1],kaus[i][0],Namen[i],Sequs[i]);
     } else {
        kaus[i][0]=b->nb[i].nachbar->nb[b->nb[i].nachbar->whichone(b)].ausrichter;
        kaus[i][1]=b->nb[i].ausrichter;
     }
   }
	
   if(b->isLeaf()) {
     vector<char> vecname;
     for(j=0;j<b->id.name.size();++j)
	 vecname.push_back(b->id.name[j]);
     vector<char> profseq;
     for(j=0;j<b->pro.size();++j)
         profseq.push_back(state_to_char(b->pro[j],typ));
     if(b->isRoot() && b->nb.size()==1) { 
        // idmat und seqmat aus Namen[0], Sequs[0] und eigenem Namen und eigener Sequenz zusammensetzen 
        for(i=0;i<Namen[0].size();++i)
            idmat.push_back(Namen[0][i]);
	idmat.push_back(vecname);
        seqmat.resize(idmat.size());
        do {
	   atnalt=atn;
           zus[0]=iter_zust(kaus[0][1],kaus[0][0],atn[0]);
	   switch(zus[0]) {
	      case START :
	      case ENDE :
	         break;
	      case PAIR :
	         for(i=0;i<Sequs[0].size();++i)
		   seqmat[i].push_back(Sequs[0][i][atnalt[0][1]]);
		 seqmat[i].push_back(profseq[atnalt[0][0]]);
	         break;
	      case DELETION :
	         for(i=0;i<Sequs[0].size();++i)
		   seqmat[i].push_back('_');
		 seqmat[i].push_back(profseq[atnalt[0][0]]);
	         break;
	      case INSERTION :
	         for(i=0;i<Sequs[0].size();++i)
		   seqmat[i].push_back(Sequs[0][i][atnalt[0][1]]);
	         seqmat[i].push_back('_');
	         break;
	      default :
	         cerr << "Unreasonable state in function push_mult\n";
		 exit(EXIT_FAILURE);
	   }
        }while(zus[0]!=ENDE);
     } else {
        // eigenen Namen in idmat und eigene Sequenz in seqmat schreiben
	// ali_to_node ist nb[0].ausrichter
        idmat.push_back(vecname);
        seqmat.push_back(profseq);
	if(b->nb.size()==1) {
           ali_to_node=kaus[0][0];
	   node_to_ali=kaus[0][1];
	} else {  // Entweder hat das Blatt mehrere Vorfahren oder es gibt nur ein Taxon
	   cerr << "Pathological situation\n";
	   exit(EXIT_FAILURE);
	}
     }
   } else {
     for(i=0;i<b->nb.size();++i)
        for(j=0;j<Namen[i].size();++j)
	   idmat.push_back(Namen[i][j]);
     seqmat.resize(idmat.size());
     int aktivins;  // index der im Moment abgearbeiteten Insertion
     int paridx=b->whichone(b->parent());  // Index des parents
     bool nendflag, syncstart;
     
     do {
       atnalt=atn;
       aktivins=-1;
       for(i=0;i<atn.size();++i) 
           zus[i]=iter_zust(kaus[i][1],kaus[i][0],atn[i]);
       nendflag=notatend(zus,syncstart);
       if(nendflag && !syncstart) {  // Wenn man das Ende noch nicht erreicht hat und nicht mehr am Anfang steht
          column.clear();
          for(i=0;i<zus.size() && (aktivins==-1 || aktivins==paridx);++i) // suche erste Insertion
              if(zus[i]==INSERTION)
	         aktivins=(int)i;
	      
          for(i=0;i<zus.size();++i)
              for(j=0;j<Sequs[i].size();++j) { // wird also fuer den parent nicht durchlaufen
	          if(aktivins>-1) {
	            if((int)i!=aktivins)
		       column.push_back('_');
                    else
		       column.push_back(Sequs[i][j][atnalt[i][1]]);
	          } else { // keine Insertionen
	               switch(zus[i]) {
		          case PAIR :
			     column.push_back(Sequs[i][j][atnalt[i][1]]);
			     break;
			  case DELETION :
			     column.push_back('_');
			     break;
                          default :
			     cerr << "Impossible state " <<(int)zus[i]<< " in push_mult\n";
			     exit(EXIT_FAILURE);   
		       }
	          }
          }
	  if(validcolumn(column,idmat.size())) {
	     for(i=0;i<column.size();++i)
	         seqmat[i].push_back(column[i]);
	     if((zus[paridx]==PAIR && aktivins==-1) || (zus[paridx]==INSERTION && aktivins==paridx))
                 //node_to_ali wird weiter unten noch laenger
                 node_to_ali.push_back(ali_to_node.size());
	     else
	         node_to_ali.push_back(ali_to_node.size()-1);
          }
	  if((zus[paridx]==PAIR && aktivins==-1) || (zus[paridx]==INSERTION && aktivins==paridx))
	    ali_to_node.push_back((int)seqmat[0].size()-1);

          for(i=0;i<atn.size();++i)
             if(aktivins>-1 && (int)i!=aktivins)  // Es gab eine Insertion, die vor den anderen Zustaenden abgearbeitet wurde
	        atn[i]=atnalt[i];
       }
     } while(nendflag);
   }
   
}

bool mdataset::validcolumn(vector<char>& col, unsigned int expsize)
{
   if(col.size()!=expsize) {
      cerr << "Alignment column has "<<col.size()<<" entries, yet there should be "<<expsize<<endl;
      exit(EXIT_FAILURE);
   }
   bool retval=TRUE;
   
   for(unsigned int i=0;i<col.size();++i)
       retval=retval && (col[i]=='_');

   return !retval;
}

bool mdataset::notatend(vector<zustand>& pzus, bool& syncstart)
{
  bool flag=TRUE;
  syncstart=TRUE;
  
  for(unsigned int i=0;i<pzus.size();++i) {
     flag=flag && (pzus[i]==ENDE);
     syncstart=syncstart && (pzus[i]==START);
     if(pzus[i]==START && syncstart==FALSE || (i>0 && syncstart==TRUE && pzus[i]!=START)) {
        cerr << "Asynchronous start in push_mult\n";
	exit(EXIT_FAILURE);
     }
  }
  
  return !flag;
}


void mdataset::row(unsigned int i, vector<char>& rowvec)  // schreibt Spalte i nach rowvec
{
   if((int)i>=laen(0)) {
      cerr << "There is no row "<<i<< " in the alignment\n";
      exit(EXIT_FAILURE);
   }
   rowvec.resize(spez());
   for(unsigned int j=0;j<spez();++j)
      rowvec[j]=Seqs[j][i];
}

/*************************************************************/
/****************** class compact_subst_pattern    ***********/
/*************************************************************/

compact_subst_pattern::compact_subst_pattern()
{
}

compact_subst_pattern::~compact_subst_pattern()
{
}

compact_subst_pattern::compact_subst_pattern(mdataset &pali)
{
   list<vector<char> > mustlist;
   vector<char> buffer;
   string s;
   unsigned int i,j;
   if(pali.spez()==0) {
      cerr << "Alignment without taxa\n";
      exit(EXIT_FAILURE);
   }
   for(i=0;i<pali.Seqs[0].size();++i) {
      pali.row(i,buffer);
      mustlist.push_back(buffer);
   }
   mustlist.sort();
   
   while(!mustlist.empty()) {
      buffer=mustlist.back();
      if(muster.empty() || muster.back()!=buffer) {
        muster.push_back(buffer);
	anzmus.push_back(1);
      } else {
        anzmus.back()+=1;
      }
      mustlist.pop_back();
   }
   
   for(i=0;i<pali.Identif.size();++i) {
      s="";
      for(j=0;j<pali.Identif[i].size();++j)
         s+=pali.Identif[i][j];
      ids.push_back(s);
   }
}

void compact_subst_pattern::operator=(compact_subst_pattern csp2)
{
   muster=csp2.muster;
   anzmus=csp2.anzmus;
   ids=csp2.ids;
}

/*************************************************************/
/****************** class proposal                 ***********/
/*************************************************************/


proposal::proposal(substmodel& sub, indel& id, keepfixed& kee)
{
  logcounter=0.0;
  logdenominator=0.0;  
  propos_subpar(sub,kee.subfix);
  propos_idpar(id,kee.indfix);
}

proposal::~proposal()
{
}


void proposal::propos_subpar(substmodel& sub, vector<bool>& subfix)
{
  vector<double> oldsubp;
  sub.paramsOf(oldsubp);
  if(oldsubp.size()!=subfix.size()) {
     cerr << "vector subfix has the wrong size\n";
     exit(EXIT_FAILURE);
  }
  subpar.resize(oldsubp.size());
  vector<bool> internalsubfix=subfix;
  if(subpar.size()>3){
     vector<int> zufpar;
     for(unsigned int i=0;i<3;++i)
        zufpar.push_back(randominteger(subpar.size()));
     for(int i=0;i<(int)internalsubfix.size();++i)
       if(i!=zufpar[0] && i!=zufpar[1] && i!=zufpar[2])
         internalsubfix[i]=TRUE;
  }
  
  for(unsigned int i=0;i<subpar.size();++i)
     if(internalsubfix[i])
        subpar[i]=oldsubp[i];
     else {
        subpar[i]=rand_exp(1.0/oldsubp[i]);
	// log des Priors
        logcounter-=subpar[i];
	logdenominator-=oldsubp[i];
	// Transition
        logcounter-=log(subpar[i])+oldsubp[i]/subpar[i];
	logdenominator-=log(oldsubp[i])+subpar[i]/oldsubp[i];
     }
}


void proposal::propos_idpar(indel& id, vector<bool>& indfix)
{
  vector<double> oldidp;
  idarch iarch=id.returnparams(oldidp);
  if(oldidp.size()!=indfix.size()) {
     cerr << "vector indfix has the wrong size\n";
     exit(EXIT_FAILURE);
  }
  switch(iarch) {
     case TKF1 :
        idpar.resize(2);
	if(!indfix[0]) {
          idpar[0]=rand_exp(1.0/oldidp[0]);// new lambda
          // log des Priors fuer lambda (logcounter und logdenominator wurden im Konstruktor =0.0 gesetzt)
	  logcounter-=idpar[0]; 
          logdenominator-=oldidp[0];
	  // Transition fuer lambda
	  logcounter-=log(idpar[0])+oldidp[0]/idpar[0]; 
	  logdenominator-=log(oldidp[0])+idpar[0]/oldidp[0];
	} else
	  idpar[0]=oldidp[0];
	if(!indfix[1]) {
          idpar[1]=idpar[0]*(1.0+1.0/rand_exp((oldidp[1]-oldidp[0])/oldidp[0])); // new mue
	  // log des Priors fuer lambda/(mue-lambda)
	  logcounter-=idpar[0]/(idpar[1]-idpar[0]);
	  logdenominator-=oldidp[0]/(oldidp[1]-oldidp[0]);
	  // Transition fuer lambda/(mue-lambda)
	  logcounter-=log(idpar[0])-log(idpar[1]-idpar[0])+oldidp[0]*(idpar[1]-idpar[0])/(idpar[0]*(oldidp[1]-oldidp[0]));
	  logdenominator-=log(oldidp[0])-log(oldidp[1]-oldidp[0])+idpar[0]*(oldidp[1]-oldidp[0])/(oldidp[0]*(idpar[1]-idpar[0]));
	} else
	  idpar[1]=oldidp[1];
        break;
     case TKF2 :
        idpar.resize(3);
	if(!indfix[0]) {
          idpar[0]=rand_exp(1.0/oldidp[0]);// new lambda
          // log des Priors fuer lambda (logcounter und logdenominator wurden im Konstruktor =0.0 gesetzt)
	  logcounter-=idpar[0]; 
          logdenominator-=oldidp[0];
	  // Transition fuer lambda
	  logcounter-=log(idpar[0])+oldidp[0]/idpar[0]; 
	  logdenominator-=log(oldidp[0])+idpar[0]/oldidp[0];
	} else
	  idpar[0]=oldidp[0];
	if(!indfix[1]) {
	  // nur zur Probe:
	  idpar[1]=idpar[0]*1.00001;
	} else
	  idpar[1]=oldidp[1];
	if(!indfix[2]) { 
	  idpar[2]=1.0/(rand_exp(oldidp[2]/(1.0-RSHIFT*oldidp[2]))+RSHIFT);
	  // log des Priors fuer 1/r-1 (Exponentialverteilung mit Erwartungswert 1)
	  logcounter-=1.0/idpar[2]-RSHIFT;
	  logdenominator-=1.0/oldidp[2]-RSHIFT;
	  // Transition fuer r
	  logcounter-=log(1.0/idpar[2]-RSHIFT)+(1.0/oldidp[2]-RSHIFT)/(1.0/idpar[2]-RSHIFT);
	  logdenominator-=log(1.0/oldidp[2]-RSHIFT)+(1.0/idpar[2]-RSHIFT)/(1.0/oldidp[2]-RSHIFT);
	} else
	  idpar[2]=oldidp[2];
        break;
     default :
        cerr << "Unknown indel model in function propos_idpar\n";
        exit(EXIT_FAILURE);
  }
}

/*************************************************************/
/****************** classless                      ***********/
/*************************************************************/

zustand get_zust(vector<int>& pAB, vector<int>& pBA, int a, int b)
{
    /* Liefert den zu a und b assoziierten Alignment-Zustand
     */
    if (a==-1) {
	if (b==-1) 
	    return START;
	if (pBA[b]==-1) 
	    return INSERTION;
	return NOSTATE;
    }
    if (a==(int)pAB.size()) {
	if (b==(int)pBA.size())
	    return ENDE;
	return NOSTATE;
    }
    if (b==-1) {
	if (pAB[a]==-1) 
	    return DELETION;
	return NOSTATE;
    }
    if (b==(int)pBA.size()) {
	return NOSTATE;
    }
    if (b==pAB[a]){
	if (a==pBA[b])
	    return PAIR;
	if (pBA[b]<a)
	    return DELETION;
	return NOSTATE;
    }
    if (pBA[b]==a && pAB[a]<b )
	return INSERTION;
    return NOSTATE;

}

zustand iter_zust(vector<int>& AvsB, vector<int>& BvsA, vector<int>& indices)
{
    zustand zust;
    zust=get_zust(AvsB,BvsA,indices[0],indices[1]);
    switch (zust) {
    case START:
    case PAIR:  
    case INSERTION: 
    case DELETION:
	if ( indices[0]==(int)AvsB.size()-1 ) {
	    ++indices[1];
	    if ( indices[1]==(int)BvsA.size() ) 
		++indices[0];
	}
	else {
	    if ( indices[1]==(int)BvsA.size()-1 )
		++indices[0];
	    else {
		if ( AvsB[indices[0]+1]==indices[1] )
		    ++indices[0];
		else {
		    ++indices[1];
		    if ( BvsA[indices[1]]!=indices[0] ) 
			++indices[0];
		}
	    }
	}
	break;
    case ENDE:
	break;
    default:
	cerr << indices[0] << " und " << indices[1] << " ergeben keinen Zustand"
	     << " in iter_zust\n";
	exit(1);
    }
    return zust;
}


bool patternequivalent(vector<char>& p1, vector<char>& p2)
// sind an den Positionsmustern p1 und p2 die gleichen Taxa beteiligt?
{
   if(p1.size()!=p2.size()) {
      cerr << "You were trying to compare position patterns with different sizes\n";
      exit(EXIT_FAILURE);
   }
   vector<char> buf1, buf2;
   unsigned int i;
   for(i=0;i<p1.size();++i) {   
      if(p1[i]=='_' || p1[i]=='-')
         buf1.push_back('_');
      else
         buf1.push_back('B');
      if(p2[i]=='_' || p2[i]=='-')
         buf2.push_back('_');
      else
         buf2.push_back('B');
   }
   return (buf1==buf2);
}

