/****************************************************************************
 * This file is part of the program ALIFRITZ                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                            		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <cfloat>  

using namespace std;

#include "seqcode.hh"
#include "data.hh"
#include "substmod.hh"
#include "indel.hh"
#include "random.hh"
#include "treespec.hh"
#include "profalign.hh"

//#define BISHIERHINZUFAELLIG 100000
#define BISHIERHINZUFAELLIG 0
#define MAXSEQLEN 1500

// Parameter fuer den 'EM'-Alg.
#define EMSAMP 200
#define EMSTOP 100
#define EMPREC 1e-2
#define MIN_TIME 0.000001
#define MAX_TIME 1000.0
#define MIN_SCALE 0.000001
#define MAX_SCALE 1000.0
// MAX_TIME = 1000 entspr. 10 Subst./Pos.
#define EPSILON 0.000001
#define MAXITS 10

#define DEBUG

/***********************************************************************************************/
/************     ein paar wichtige Funktionen             *************************************/
/***********************************************************************************************/


double addlog(double a, double b) {
    // liefert log(exp(a)+exp(b))
    if (a>b) return a+log(1+exp(b-a));
    else return b+log(1+exp(a-b));
}

double addlog(double a, double b, double c) {
    // liefert log(exp(a)+exp(b)+exp(c))
    return addlog(addlog(a,b),c);
}

double addlog(vector<double>& v) {
    // liefert log(Summe_ueber_i von exp(v[i]))
    if(v.size()==0) {
       cerr << "Not enough arguments to addlog\n";
       exit(EXIT_FAILURE);
    }
    if(v.size()==1)
       return v[0];
    if(v.size()==2)
       return addlog(v[0],v[1]);
    vector<double> v2 = v;
    v2.pop_back();
    return addlog(addlog(v2),v.back());
}

double maximum(double a, double b) {
    return ( a>b ? a : b );
}

double maximum(double a, double b, double c) {
    return maximum(maximum(a,b),c);
}

int maximum(int a, int b) {
    return ( a>b ? a : b );
}

// distrib_pow ersetzt die Werte in distrib durch deren expo hoch itePot Potenz und renormiert
void distrib_pow(vector<double>& distrib, double expo, unsigned long int itePot)
{
   double sum;
   unsigned long int j;

/*  viel zu langsam 
   for(i=0;i<=itePot;++i) {
      sum = 0.0;
      if(i==0) {
        for(j=0;j<distrib.size();++j) {
	    if(distrib[j]<0.0) {
	       cerr << "One of the frequencies in distrib_pow is negative!\n";
	       exit(EXIT_FAILURE);
	    }
	    lokal.push_back(distrib[j]);
	    sum += distrib[j];
	}
      } else {
        
        for(j=0;j<lokal.size();++j) {
	    lokal[j]=pow(lokal[j],expo);
	    sum += lokal[j];
	}
      }
      if(sum < 0.0+DBL_MIN) {
	 cerr << "The probability vector that was passed to distrib_pow is empty!\n";
	 exit(EXIT_FAILURE); 
      } else   // Renormierung in jedem Schritt
         for(j=0;j<lokal.size();++j)
	     lokal[j]/=sum;
   }*/
   
   sum = 0.0;
   if(itePot==0) {
        for(j=0;j<distrib.size();++j) {
	    if(distrib[j]<0.0) {
	       cerr << "One of the frequencies in distrib_pow is negative!\n";
	       exit(EXIT_FAILURE);
	    }
	    sum += distrib[j];
	}
        if(sum < 0.0+DBL_MIN) {
	   cerr << "The probability vector that was passed to distrib_pow is empty!\n";
	   exit(EXIT_FAILURE); 
        }
        for(j=0;j<distrib.size();++j)
            distrib[j]/=sum;
   } else {
        double d=pow(expo,(double)itePot);
        if(d<10.0) {
            for(j=0;j<distrib.size();++j) {
	        distrib[j]=pow(distrib[j],d);
	        sum += distrib[j];
	    }
            if(sum < 0.0+DBL_MIN) {
	       cerr << "The probability vector that was passed to distrib_pow is empty!\n";
	       exit(EXIT_FAILURE); 
            }
            for(j=0;j<distrib.size();++j)
                distrib[j]/=sum;
        } else { // Exaktheit ist hier nicht mehr wichtig
            unsigned int idx=0;
	    double max=-1;
	    for(j=0;j<distrib.size();++j)
	        if(distrib[j]>max) {
		   idx=j;
		   max=distrib[j];
		}
	    for(j=0;j<distrib.size();++j)
	        if(j==idx)
		   distrib[j]=1.0;
		else
		   distrib[j]=0.0;
        }
   }
}


/***********************************************************************************************/
/************     class align_graph                                    *************************/
/***********************************************************************************************/

//  private :

void align_graph::beschrifte()
{
    zustand firststate = algn->get_zust(von_A,von_B),
	laststate  = algn->get_zust(bis_A,bis_B);
    beschrifte(firststate,laststate);
}

void align_graph::beschrifte(zustand firststate, zustand laststate)
{
    int v_A=von_A+1,v_B=von_B+1,b_A=0,b_B=0,i,j;
    switch (laststate) {
    case ENDE :
	b_A=algn->pro_A->laen()-1;
	b_B=algn->pro_B->laen()-1;
	break;
    case PAIR :
	b_A=bis_A-1;
	b_B=bis_B-1;
	break;
    case INSERTION :
	b_A=bis_A;
	b_B=bis_B-1;
	break;
    case DELETION :
	b_A=bis_A-1;
	b_B=bis_B;
	break;
    default:
	cerr << "komischer laststate in align_graph::beschrifte(...)\n";
    } 
    
    double logUewk[4][5];
    for(i=0;i<4;++i)
       for(j=0;j<5;++j)
          logUewk[i][j]=algn->idm->retlnprob(i,j);

    A_gap_kante[0][0]=logUewk[firststate][INSERTION]+algn->sm->logfreqofstate((*algn->pro_B)[v_B]);
    B_gap_kante[0][0]=logUewk[firststate][DELETION]+algn->sm->logfreqofstate((*algn->pro_A)[v_A]);
    paar_kante[0][0]=logUewk[firststate][PAIR]+algn->sm->logfreqofstate((*algn->pro_A)[v_A])
                                              +algn->sm->logProb_ij((*algn->pro_A)[v_A],(*algn->pro_B)[v_B]);

    for(i=1;i<=b_A-v_A;++i) {
	A_gap_kante[i][0]=B_gap_kante[i-1][0]+logUewk[DELETION][INSERTION]
	                 +algn->sm->logfreqofstate((*algn->pro_B)[v_B]);
	B_gap_kante[i][0]=B_gap_kante[i-1][0]+logUewk[DELETION][DELETION]
	                 +algn->sm->logfreqofstate((*algn->pro_A)[v_A+i]);
	paar_kante[i][0]=B_gap_kante[i-1][0]+logUewk[DELETION][PAIR]
	                 +algn->sm->logfreqofstate((*algn->pro_A)[v_A+i])
	                 +algn->sm->logProb_ij((*algn->pro_A)[i+v_A],(*algn->pro_B)[v_B]);
    }
    A_gap_kante[b_A-v_A+1][0]=B_gap_kante[b_A-v_A][0]+logUewk[DELETION][INSERTION]
                             +algn->sm->logfreqofstate((*algn->pro_B)[v_B]);

    for(j=1;j<=b_B-v_B;++j) {
	B_gap_kante[0][j]=A_gap_kante[0][j-1]+logUewk[INSERTION][DELETION]
	                 +algn->sm->logfreqofstate((*algn->pro_A)[v_A]);
	A_gap_kante[0][j]=A_gap_kante[0][j-1]+logUewk[INSERTION][INSERTION]
	                 +algn->sm->logfreqofstate((*algn->pro_B)[v_B+j]);
	paar_kante[0][j]=A_gap_kante[0][j-1]+logUewk[INSERTION][PAIR]
	                 +algn->sm->logfreqofstate((*algn->pro_A)[v_A])
	                 +algn->sm->logProb_ij((*algn->pro_A)[v_A],(*algn->pro_B)[v_B+j]);
    }
    B_gap_kante[0][b_B-v_B+1]=A_gap_kante[0][b_B-v_B]+logUewk[INSERTION][DELETION]
                             +algn->sm->logfreqofstate((*algn->pro_A)[v_A]);



    for(i=1;i<=b_A-v_A;++i) {
	for(j=1;j<=b_B-v_B;++j) {
	    A_gap_kante[i][j]=addlog(B_gap_kante[i-1][j]+logUewk[DELETION][INSERTION],
			             paar_kante[i-1][j-1]+logUewk[PAIR][INSERTION],
			             A_gap_kante[i][j-1]+logUewk[INSERTION][INSERTION])
			     +algn->sm->logfreqofstate((*algn->pro_B)[v_B+j]);
	    B_gap_kante[i][j]=addlog(B_gap_kante[i-1][j]+logUewk[DELETION][DELETION],
			             paar_kante[i-1][j-1]+logUewk[PAIR][DELETION],
			             A_gap_kante[i][j-1]+logUewk[INSERTION][DELETION])
			     +algn->sm->logfreqofstate((*algn->pro_A)[v_A+i]);
	    paar_kante[i][j]=addlog(B_gap_kante[i-1][j]+logUewk[DELETION][PAIR],
			            paar_kante[i-1][j-1]+logUewk[PAIR][PAIR],
			            A_gap_kante[i][j-1]+logUewk[INSERTION][PAIR])
			    +algn->sm->logfreqofstate((*algn->pro_A)[v_A+i])
			    +algn->sm->logProb_ij((*algn->pro_A)[v_A+i],(*algn->pro_B)[v_B+j]);
	}
    }
    for(i=1;i<=b_A-v_A;++i)
	B_gap_kante[i][b_B-v_B+1]=addlog(B_gap_kante[i-1][b_B-v_B+1]+logUewk[DELETION][DELETION],
		                         paar_kante[i-1][b_B-v_B]+logUewk[PAIR][DELETION],
		                         A_gap_kante[i][b_B-v_B]+logUewk[INSERTION][DELETION])
			         +algn->sm->logfreqofstate((*algn->pro_A)[v_A+i]);
    for(j=1;j<=b_B-v_B;++j)
	A_gap_kante[b_A-v_A+1][j]=addlog(B_gap_kante[b_A-v_A][j]+logUewk[DELETION][INSERTION],
		                         paar_kante[b_A-v_A][j-1]+logUewk[PAIR][INSERTION],
		                         A_gap_kante[b_A-v_A+1][j-1]+logUewk[INSERTION][INSERTION])
				 +algn->sm->logfreqofstate((*algn->pro_B)[v_B+j]);
}


//  public :

align_graph::align_graph(const alignment& ali,
			 int v_A,int v_B,int b_A,int b_B)
    /* Konstruiere einen Alignment-Graphen fuer das Teilalignment zwischen
   * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
   * von algn
   */
    : A_gap_kante(b_A-v_A+2,vector<double>(b_B-v_B+1,DBL_MAX)) ,
      B_gap_kante(b_A-v_A+1,vector<double>(b_B-v_B+2,DBL_MAX)) ,
      paar_kante(b_A-v_A+1,vector<double>(b_B-v_B+1,DBL_MAX))
{
    algn = &ali;
    von_A=v_A; bis_A=b_A; von_B=v_B; bis_B=b_B;
    beschrifte();
}

align_graph::align_graph(const alignment& ali,
			 int v_A,int v_B,int b_A,int b_B, zustand firststate, zustand laststate)
    /* Konstruiere einen Alignment-Graphen fuer das Teilalignment zwischen
   * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
   * von algn
   */
    : A_gap_kante(b_A-v_A+2,vector<double>(b_B-v_B+1,DBL_MAX)) ,
      B_gap_kante(b_A-v_A+1,vector<double>(b_B-v_B+2,DBL_MAX)) ,
      paar_kante(b_A-v_A+1,vector<double>(b_B-v_B+1,DBL_MAX))
{
    algn = &ali;
    von_A=v_A; bis_A=b_A; von_B=v_B; bis_B=b_B;
    beschrifte(firststate,laststate);
}

align_graph::~align_graph()
{
}

zustand align_graph::zufallsschritt(zustand nextstep, int a, int b, const int PA,  const int PB,
                                    double expo, unsigned long int itePot) const
    /* Gibt einen zufaelligen Zustand aus, der dem Schritt hin zu Knoten
     * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
     * es von (a,b) aus mit einem Schritt von Zustand nextstep weitergeht.
     * Durch expo und itePot wird diese Verteilung "fokussiert"
     */
{
    a-=von_A+1; b-=von_B+1;
    if ( a<0 || b<0 ) 
	cerr << "Fehler: zu kleine Parameter in align_graph::zufallsschritt!!!\n";
    if (a<=0) {
	if (b<=0) cerr << "Fehler: a,b<=0 in  align_graph::zufallsschritt !!!\n";
	//cout << A_GAP <<endl;
	return INSERTION;
    }
    if (b==0) {
	//cout << B_GAP << endl;
	return DELETION;
    }
    double basprob;
    switch (nextstep) {
        case ENDE:
	   basprob=0.0;
	   break;
	case PAIR:
	   basprob=algn->sm->logfreqofstate(PA)+algn->sm->logProb_ij(PA,PB);
	   break;
	case INSERTION:
	   basprob=algn->sm->logfreqofstate(PB);
	   break;
	case DELETION:
	   basprob=algn->sm->logfreqofstate(PA);
	   break;
	default:
	   cerr << "Unreasonable state nextstep\n";
	   exit(EXIT_FAILURE);
    }

    vector<double> distrib;
    double R=randomunitintervall();
    double P_Paar=paar_kante[a-1][b-1]+algn->idm->retlnprob(PAIR,nextstep)+basprob,
           P_A_GAP=A_gap_kante[a][b-1]+algn->idm->retlnprob(INSERTION,nextstep)+basprob,
           P_B_GAP=B_gap_kante[a-1][b]+algn->idm->retlnprob(DELETION,nextstep)+basprob,
	   lsumme=addlog(P_Paar,P_A_GAP,P_B_GAP); 
    distrib.push_back(exp(P_Paar-lsumme));distrib.push_back(exp(P_A_GAP-lsumme));distrib.push_back(exp(P_B_GAP-lsumme));
    distrib_pow(distrib,expo,itePot);

    if ( R < distrib[0] ) {
	//cout << A_GAP << endl;
	return PAIR;
    }
    if ( R < distrib[0]+distrib[1] ) {
	//cout << B_GAP << endl;
	return INSERTION;
    }
    //cout << PAARUNG << endl;
    return DELETION;
}  

zustand align_graph::maxwk_schritt(zustand nextstep, int a, int b, const int PA,  const int PB) const
    /* Gibt einen wahrscheinlichsten Zustand aus, der dem Schritt hin zu Knoten
     * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
     * es von (a,b) aus mit einem Schritt von Zustand nextstep weitergeht.
     * Durch expo und itePot wird diese Verteilung "fokussiert"
     */
{
    a-=von_A+1; b-=von_B+1;
    if ( a<0 || b<0 ) 
	cerr << "Fehler: zu kleine Parameter in align_graph::zufallsschritt!!!\n";
    if (a<=0) {
	if (b<=0) cerr << "Fehler: a,b<=0 in  align_graph::zufallsschritt !!!\n";
	//cout << A_GAP <<endl;
	return INSERTION;
    }
    if (b==0) {
	//cout << B_GAP << endl;
	return DELETION;
    }
    double basprob;
    switch (nextstep) {
        case ENDE:
	   basprob=0.0;
	   break;
	case PAIR:
	   basprob=algn->sm->logfreqofstate(PA)+algn->sm->logProb_ij(PA,PB);
	   break;
	case INSERTION:
	   basprob=algn->sm->logfreqofstate(PB);
	   break;
	case DELETION:
	   basprob=algn->sm->logfreqofstate(PA);
	   break;
	default:
	   cerr << "Unreasonable state nextstep\n";
	   exit(EXIT_FAILURE);
    }

    double l_Paar=paar_kante[a-1][b-1]+algn->idm->retlnprob(PAIR,nextstep)+basprob,
           l_A_GAP=A_gap_kante[a][b-1]+algn->idm->retlnprob(INSERTION,nextstep)+basprob,
           l_B_GAP=B_gap_kante[a-1][b]+algn->idm->retlnprob(DELETION,nextstep)+basprob; 

    if ( l_B_GAP < l_A_GAP ) { 
	if( l_Paar <  l_A_GAP )
	    return INSERTION;
	else
	    return PAIR;
    }
    else {
	if( l_Paar <  l_B_GAP )
	    return DELETION;
	else
	    return PAIR;
    }
}  

/***********************************************************************************************/
/************     class alignment                                      *************************/
/***********************************************************************************************/

alignment::alignment()
/* Gibt eine Fehlermeldung */
{
   cerr << "You were trying to construct an empty alignment!\n";
   exit(EXIT_FAILURE);
}

alignment::alignment(const bioseq &p_1,const bioseq &p_2, double time,
	      const substmodel &sub,const indel &ind,
	      double coolali, unsigned long int step)
/* Initialisiert gemaess Substitutionsmodell sub, Indelmodell ind
   und unter Beruecksichtigung des SimAnneal-Schritts step */
{
   pro_A = &p_1;
   pro_B = &p_2;
   pAB.resize(pro_A->size());
   pBA.resize(pro_B->size());
   sm = new substmodel(sub);
   sm->settime(time);
   idm = new indel(ind);
   idm->fillP(time);
   coolexp = coolali;
   Schritt = step;
   if(pro_A->size()==0) { // Sequenz A ist leer
     for(unsigned int i=0;i<pBA.size();++i)
         pBA[i]=-1;
   } else {
      if(pro_B->size()==0) { // Sequenz B ist leer
         for(unsigned int i=0;i<pAB.size();++i)
             pAB[i]=-1;
      } else
          sample(); // sampelt ein Alignment
   }
}   
   
alignment::alignment(const bioseq &p_1,const bioseq &p_2, double time,
	      const substmodel &sub,const indel &ind)
/* Konstruiert nicht wirklich ein Alignment, aber bereitet alles
 * fuer den 'EM'-Alg. vor. Also Vorsicht!
 */
{
   // 'Dummy' constructor which should only be used by experts!!!!!

   pro_A = &p_1;
   pro_B = &p_2;
   pAB.resize(pro_A->size());
   pBA.resize(pro_B->size());
   for(unsigned int i=0;i<pBA.size();++i)
       pBA[i]=-1;
   for(unsigned int i=0;i<pAB.size();++i)
       pAB[i]=-1;
   sm = new substmodel(sub);
   sm->settime(time);
   idm = new indel(ind);
   idm->fillP(time);
   Schritt = 0;
}   
   
alignment::alignment(const baum &b_1,const baum &b_2,
	      const substmodel &sub,const indel &ind,
	      double coolali, unsigned long int step)
/* Initialisiert gemaess dem Alignment der Sequenzen an b_1
   und b_2, dem Substitutionsmodell sub, dem Indelmodell ind
   und unter Beruecksichtigung des SimAnneal-Schritts step */
{
   // sind b_1 und b_2 Nachbarn
   int idxA = b_1.whichone(&b_2);
   int idxB = b_2.whichone(&b_1);
   
   // zuweisen
   pro_A = &(b_1.pro);
   pro_B = &(b_2.pro);
   pAB = b_1.nb[idxA].ausrichter;
   pBA = b_2.nb[idxB].ausrichter;
   double time = b_1.nb[idxA].length;
   sm = new substmodel(sub);
   sm->settime(time);
   idm = new indel(ind);
   idm->fillP(time);
   coolexp = coolali;
   Schritt = step;
   
   if(pBA.size()==0 || pAB.size()==0) {
      if(pBA.size()!=pro_B->size() || pAB.size()!=pro_A->size())
      {
          cerr << "Found inconsistencies while constructing an alignment!\n"
               << pBA.size() << " " << pro_B->size() <<endl 
               << pAB.size() << " " << pro_A->size() <<endl; 
          exit(EXIT_FAILURE);
      } else 
          cerr << "Warning:there are internal sequences with length 0\n";
   } else {
      // stimmt auch alles?
      if(pBA.size()!=pro_B->size() || pAB.size()!=pro_A->size() ||
        ((pBA.back()!=pro_A->laen()-1 || pAB.back()>pro_B->laen()-1) && 
         (pAB.back()!=pro_B->laen()-1 || pBA.back()>pro_A->laen()-1)))
      {
          cerr << "Found inconsistencies while constructing an alignment!\n"
               << pBA.size()-1 << " " << pro_B->size()-1 << " " << pAB.back()<<endl 
               << pAB.size()-1 << " " << pro_A->size()-1 << " " << pBA.back()<<endl; 
          exit(EXIT_FAILURE);
      }
   }
}   
   
alignment::alignment(const alignment& align)
/* Copy-Kostruktor */
{
   pro_A = align.pro_A;
   pro_B = align.pro_B;
   pAB = align.pAB;
   pBA = align.pBA;
   sm = new substmodel(*(align.sm));
   idm = new indel(*(align.idm));
   coolexp = align.coolexp;
   Schritt = align.Schritt;
}

alignment::~alignment()
{
   delete(sm);
   delete(idm);
}


void alignment::sample()
// sampelt einen ganzen Alignmentpfad
{
   teilsample(-1,-1,pro_A->laen(),pro_B->laen());
}


void alignment::hirschberg(bool forEM)
{
    vector<zustand> gesamtvec;
    hirschberg(-1,-1,pro_A->laen(),pro_B->laen(),gesamtvec,START,ENDE);

    int aidx=-1, bidx=-1;
    for(unsigned int i=0;i<gesamtvec.size();++i) {
        switch(gesamtvec[i]) {
            case PAIR :
	      ++aidx; ++bidx;
	      pAB[aidx]=bidx; pBA[bidx]=aidx;
	      break;
	    case INSERTION :
	      ++bidx;
	      pBA[bidx]=aidx;
	      break;
	    case DELETION :
	      ++aidx;
	      pAB[aidx]=bidx;
	      break;
	    default :
	      cerr << "Impossible state in Hirschberg algorithm\n";
	      exit(EXIT_FAILURE);
        }
    }//cout<<aidx<<" "<<bidx<<endl;

    if(forEM) { // hmmtrans und basep vorbereiten und fuellen
      hmmtrans.resize(4);
      for(unsigned int j=0;j<hmmtrans.size();++j)
          hmmtrans[j].resize(5);
      alphsize alf=sm->retalf();
      basep.resize(alf);
      for(unsigned int j=0;j<basep.size();++j)
          basep[j].resize(alf);
      for(unsigned int j=0;j<hmmtrans.size();++j)
          for(unsigned int k=0;k<hmmtrans[j].size();++k)
            // if(j==START || k==ENDE)
	        hmmtrans[j][k]=0.0;
	    // else
	      //  hmmtrans[j][k]=1.0; // einfacher Pseudocount
      for(int j=0;j<alf;++j)
         for(int k=0;k<alf;++k)
	     basep[j][k]=0.0;
	     //basep[j][k]=1.0;   // einfacher Pseudocount
      int i=-1; int j=-1;
      int oldi=i, oldj=j;
      int PA=-1, PB=-1;
      zustand zus=iter_zust(&i,&j), zusalt=START;
      while(zus!=ENDE) {
         zusalt=zus;
         zus=iter_zust(&i,&j);
         hmmtrans[zusalt][zus]+=1.0;
         switch (zus) {
	       case PAIR :
	            PA=(*pro_A)[oldi]; PB=(*pro_B)[oldj];
                    // aus Faulheit wird im folgenden nicht
		    // zwischen den verschiedenen Wildcards unterschieden!!!!
	            if(PA<alf) {
	               if(PB<alf) {
		         basep[PA][PB]+=1.0;
		       } else {
		         for(int k=0;k<alf;++k)
		           basep[PA][k]+=1.0/(double)alf;
		       }
		    } else {
	               if(PB<alf) {
		         for(int k=0;k<alf;++k)
		             basep[k][PB]+=1.0/(double)alf;
		       } else {
		         for(int k=0;k<alf;++k)
		           for(int l=0;l<alf;++l)
		             basep[k][l]+=1.0/(double)(alf*alf);
		       }
	            }
		    break;
	       case INSERTION : PA=-1; PB=(*pro_B)[oldj]; break;
	       case DELETION : PA=(*pro_A)[oldi]; PB=-1; break;
	       case ENDE : break;
	       default: cerr << "unmoeglicher \"nextstate\" "<<(int)zus<<" in opt_for_giv!!!\n";
	                exit(EXIT_FAILURE);
          }
          oldi=i;oldj=j;
      }
    }
}

void alignment::hirschberg(int von_A, int von_B, int bis_A, int bis_B, vector<zustand>& gesamtvec, zustand firststate, zustand laststate)
{
///cout<<von_A<<" "<<bis_A<<" "<<von_B<<" "<<bis_B<<endl;
    gesamtvec.clear();
    int leftA=von_A+1, leftB=von_B+1, rightA=bis_A, rightB=bis_B;
    switch(laststate) {
       case PAIR:
       case ENDE:
          --rightA;
	  --rightB;
	  break;
       case INSERTION:
          --rightB;
	  break;
       case DELETION:
          --rightA;
	  break;
       default:
          cerr << "Unreasonable last state in hirschberg()\n";
	  exit(EXIT_FAILURE);
    }
    int oldrightA=rightA, oldrightB=rightB;
    double logUewk[4][5];
    for(int i=0;i<4;++i)
       for(int j=0;j<5;++j)
          logUewk[i][j]=idm->retlnprob(i,j);

    if(rightA-leftA>=3 && rightB>=leftB) {
       int cut=(rightA+leftA)/2;
       vector<vector<double> > fPair(2,vector<double>(rightB-leftB+1,DBL_MAX)),
                               fDel(2,vector<double>(rightB-leftB+2,DBL_MAX)),
			       fIns(3,vector<double>(rightB-leftB+1,DBL_MAX)),
			       bPair(2,vector<double>(rightB-leftB+1,DBL_MAX)),
			       bIns(2,vector<double>(rightB-leftB+1,DBL_MAX)),
			       bDel(2,vector<double>(rightB-leftB+1,DBL_MAX));

       // FORWARD
       // zweite Zeile wird ausgefuellt, da diese spaeter die erste ueberschreibt
       fIns[1][0]=logUewk[firststate][INSERTION]+sm->logfreqofstate((*pro_B)[leftB]);
       fDel[1][0]=logUewk[firststate][DELETION]+sm->logfreqofstate((*pro_A)[leftA]);
       fPair[1][0]=logUewk[firststate][PAIR]+sm->logfreqofstate((*pro_A)[leftA])
                   +sm->logProb_ij((*pro_A)[leftA],(*pro_B)[leftB]);

       for(int j=1;j<=rightB-leftB;++j) {
	   fDel[1][j]=fIns[1][j-1]+logUewk[INSERTION][DELETION]
	             +sm->logfreqofstate((*pro_A)[leftA]);
	   fIns[1][j]=fIns[1][j-1]+logUewk[INSERTION][INSERTION]
	             +sm->logfreqofstate((*pro_B)[leftB+j]);
	   fPair[1][j]=fIns[1][j-1]+logUewk[INSERTION][PAIR]
	              +sm->logfreqofstate((*pro_A)[leftA])
	              +sm->logProb_ij((*pro_A)[leftA],(*pro_B)[leftB+j]);
       }
       fDel[1][rightB-leftB+1]=fIns[1][rightB-leftB]+logUewk[INSERTION][DELETION]
                              +sm->logfreqofstate((*pro_A)[leftA]);
       fIns[2][0]=fDel[1][0]+logUewk[DELETION][INSERTION]+sm->logfreqofstate((*pro_B)[leftB]);
       for(int j=1;j<=rightB-leftB;++j) {
	       fIns[2][j]=addlog(fDel[1][j]+logUewk[DELETION][INSERTION],
			         fPair[1][j-1]+logUewk[PAIR][INSERTION],
			         fIns[2][j-1]+logUewk[INSERTION][INSERTION])
	                  +sm->logfreqofstate((*pro_B)[leftB+j]);
       }
       while(leftA<cut) {
           fDel[0]=fDel[1];fIns[0]=fIns[1];fIns[1]=fIns[2];fPair[0]=fPair[1];
	   fIns[1][0]=fDel[0][0]+logUewk[DELETION][INSERTION]
	              +sm->logfreqofstate((*pro_B)[leftB]);
	   fDel[1][0]=fDel[0][0]+logUewk[DELETION][DELETION]
	              +sm->logfreqofstate((*pro_A)[leftA+1]);
	   fPair[1][0]=fDel[0][0]+logUewk[DELETION][PAIR]
	               +sm->logfreqofstate((*pro_A)[leftA+1])
	               +sm->logProb_ij((*pro_A)[leftA+1],(*pro_B)[leftB]);
           for(int j=1;j<=rightB-leftB;++j) {
	       fIns[1][j]=addlog(fDel[0][j]+logUewk[DELETION][INSERTION],
			         fPair[0][j-1]+logUewk[PAIR][INSERTION],
			         fIns[1][j-1]+logUewk[INSERTION][INSERTION])
	                  +sm->logfreqofstate((*pro_B)[leftB+j]);
	       fDel[1][j]=addlog(fDel[0][j]+logUewk[DELETION][DELETION],
			         fPair[0][j-1]+logUewk[PAIR][DELETION],
			         fIns[1][j-1]+logUewk[INSERTION][DELETION])
		          +sm->logfreqofstate((*pro_A)[leftA+1]);
	       fPair[1][j]=addlog(fDel[0][j]+logUewk[DELETION][PAIR],
			          fPair[0][j-1]+logUewk[PAIR][PAIR],
			          fIns[1][j-1]+logUewk[INSERTION][PAIR])
			   +sm->logfreqofstate((*pro_A)[leftA+1])
			   +sm->logProb_ij((*pro_A)[leftA+1],(*pro_B)[leftB+j]);
           }
	   fDel[1][rightB-leftB+1]=addlog(fDel[0][rightB-leftB+1]+logUewk[DELETION][DELETION],
		                          fPair[0][rightB-leftB]+logUewk[PAIR][DELETION],
		                          fIns[1][rightB-leftB]+logUewk[INSERTION][DELETION])
			           +sm->logfreqofstate((*pro_A)[leftA+1]);
           fIns[2][0]=fDel[1][0]+logUewk[DELETION][INSERTION]+sm->logfreqofstate((*pro_B)[leftB]);
           for(int j=1;j<=rightB-leftB;++j) {
	           fIns[2][j]=addlog(fDel[1][j]+logUewk[DELETION][INSERTION],
			             fPair[1][j-1]+logUewk[PAIR][INSERTION],
			             fIns[2][j-1]+logUewk[INSERTION][INSERTION])
	                      +sm->logfreqofstate((*pro_B)[leftB+j]);
           }
	   ++leftA;
       }

       // BACKWARD
       // zweite Zeile wird ausgefuellt, da diese spaeter die erste ueberschreibt
       if(laststate==ENDE) {
          bIns[1][0]=logUewk[INSERTION][ENDE];
          bDel[1][0]=logUewk[DELETION][ENDE];
          bPair[1][0]=logUewk[PAIR][ENDE];
       } else {
          bIns[1][0]=bDel[1][0]=bPair[1][0]=0;
       }
       for(int j=1;j<=rightB-leftB;++j) {
	   bDel[1][j]=bIns[1][j-1]+logUewk[DELETION][INSERTION]
	             +sm->logfreqofstate((*pro_B)[rightB-j+1]);
	   bIns[1][j]=bIns[1][j-1]+logUewk[INSERTION][INSERTION]
	             +sm->logfreqofstate((*pro_B)[rightB-j+1]);
	   bPair[1][j]=bIns[1][j-1]+logUewk[PAIR][INSERTION]
	              +sm->logfreqofstate((*pro_B)[rightB-j+1]);
       }
       while(rightA>cut) {
           bDel[0]=bDel[1];bIns[0]=bIns[1];bPair[0]=bPair[1];
	   bIns[1][0]=bDel[0][0]+logUewk[INSERTION][DELETION]
	              +sm->logfreqofstate((*pro_A)[rightA]);
	   bDel[1][0]=bDel[0][0]+logUewk[DELETION][DELETION]
	              +sm->logfreqofstate((*pro_A)[rightA]);
	   bPair[1][0]=bDel[0][0]+logUewk[PAIR][DELETION]
	               +sm->logfreqofstate((*pro_A)[rightA]);
           for(int j=1;j<=rightB-leftB;++j) {
	       bIns[1][j]=addlog(bDel[0][j]+logUewk[INSERTION][DELETION]+sm->logfreqofstate((*pro_A)[rightA]),
			         bPair[0][j-1]+logUewk[INSERTION][PAIR]+sm->logfreqofstate((*pro_A)[rightA])+sm->logProb_ij((*pro_A)[rightA],(*pro_B)[rightB-j+1]),
			         bIns[1][j-1]+logUewk[INSERTION][INSERTION]+sm->logfreqofstate((*pro_B)[rightB-j+1]));
	       bDel[1][j]=addlog(bDel[0][j]+logUewk[DELETION][DELETION]+sm->logfreqofstate((*pro_A)[rightA]),
			         bPair[0][j-1]+logUewk[DELETION][PAIR]+sm->logfreqofstate((*pro_A)[rightA])+sm->logProb_ij((*pro_A)[rightA],(*pro_B)[rightB-j+1]),
			         bIns[1][j-1]+logUewk[DELETION][INSERTION]+sm->logfreqofstate((*pro_B)[rightB-j+1]));
	       bPair[1][j]=addlog(bDel[0][j]+logUewk[PAIR][DELETION]+sm->logfreqofstate((*pro_A)[rightA]),
			          bPair[0][j-1]+logUewk[PAIR][PAIR]+sm->logfreqofstate((*pro_A)[rightA])+sm->logProb_ij((*pro_A)[rightA],(*pro_B)[rightB-j+1]),
			          bIns[1][j-1]+logUewk[PAIR][INSERTION]+sm->logfreqofstate((*pro_B)[rightB-j+1]));
	   }
           --rightA;
       }

       vector<double> durchIns(fIns[2].size()), durchDel(bDel[1].size()), durchPair(fPair[1].size());
       for(unsigned int i=0;i<durchIns.size();++i)
           durchIns[i]=fIns[2][i]+bIns[1][bIns[1].size()-1-i];
       for(unsigned int i=0;i<durchPair.size();++i)
           durchPair[i]=fPair[1][i]+bPair[1][bPair[1].size()-1-i];
       for(unsigned int i=0;i<durchDel.size();++i)
           durchDel[i]=fDel[1][i+1]+bDel[1][bDel[1].size()-1-i]; // Vorsicht: Index von durchDel ist um -1 verschoben!!!

       vector<zustand> leftvec, rightvec, lokvec;
       zustand bz=NOSTATE;
       unsigned int maxidx=0;
       double max=-DBL_MAX;
   //    vector<double> *forvec=NULL, *backvec=NULL;
       for(unsigned int i=0;i<durchPair.size();++i) {
           if(durchPair[i]>max) {
              bz=PAIR;
	      maxidx=i;
	      max=durchPair[i];
	   }
           if(durchIns[i]>max) {
              bz=INSERTION;
	      maxidx=i;
	      max=durchIns[i];
	   }
           if(durchDel[i]>max) {
              bz=DELETION;
	      maxidx=i;
	      max=durchDel[i];
	   }
        }

	int idx=maxidx;
        rightB=leftB+maxidx;//cout<<"Schnitt: "<<leftA<<" "<<leftB+maxidx<<endl;
        int newA=leftA-1;
        zustand newz=bz, oldz=bz;//cout<<"f:"<<(int)oldz<<" ";
        while(leftA>newA || oldz==INSERTION) {
	   if(idx==0 && oldz!=DELETION) {
              newz=DELETION;
	      if(oldz==PAIR) {
	         --leftA;
		 --idx;
	      } else {
                 --idx;
                 ++newA; //Pfui!!!
	      }
	   } else {
               switch(oldz) {
                  case PAIR:
	             if(fDel[0][idx]+logUewk[DELETION][PAIR]>fPair[0][idx-1]+logUewk[PAIR][PAIR]) {
                       if(fIns[1][idx-1]+logUewk[INSERTION][PAIR]>fDel[0][idx]+logUewk[DELETION][PAIR]) {
                         newz=INSERTION;
		       } else {
                         newz=DELETION;
		       }
	             } else {
                       if(fIns[1][idx-1]+logUewk[INSERTION][PAIR]>fPair[0][idx-1]+logUewk[PAIR][PAIR]) {
                         newz=INSERTION;
		       } else {
                         newz=PAIR;
		       }
	             }
		     --leftA;--idx;
	             break;
                  case INSERTION:
	             if(fDel[1][idx]+logUewk[DELETION][INSERTION]>fPair[1][idx-1]+logUewk[PAIR][INSERTION]) {
	                if(fIns[2][idx-1]+logUewk[INSERTION][INSERTION]>fDel[1][idx]+logUewk[DELETION][INSERTION]) {
                           newz=INSERTION;
		        } else {
                           newz=DELETION;
		        }
	             } else {
                        if(fIns[2][idx-1]+logUewk[INSERTION][INSERTION]>fPair[1][idx-1]+logUewk[PAIR][INSERTION]) {
                           newz=INSERTION;
		        } else {
                           newz=PAIR;
		        }
	             }
		     --idx;
                     break;
	          case DELETION:
                     if(fDel[0][idx+1]+logUewk[DELETION][DELETION]>fPair[0][idx]+logUewk[PAIR][DELETION]) {
                       if(fIns[1][idx]+logUewk[INSERTION][DELETION]>fDel[0][idx+1]+logUewk[DELETION][DELETION]) {
                         newz=INSERTION;
		       } else {
                         newz=DELETION;
		       }
	             } else {
                       if(fIns[1][idx]+logUewk[INSERTION][DELETION]>fPair[0][idx]+logUewk[PAIR][DELETION]) {
                         newz=INSERTION;
		       } else {
                         newz=PAIR;
		       }
	             }
		     --leftA;
	             break;
                  default :
                     cerr << "Impossible state in hirschberg I\n";
	             exit(EXIT_FAILURE);
               }
	   }
	   oldz=newz;
	   leftvec.push_back(newz);//cout<<(int)newz<<" ";
        }//cout<<endl;
        leftB+=idx;//cout<<"Left:"<<leftA<<" "<<leftB<<endl;
        newA=rightA+1;
        newz=oldz=bz;
        while(rightA<-maximum(-newA,-oldrightA)) {//cout<<"b:";
             if(oldrightB-rightB==0) {
                  newz=DELETION;
		  ++rightA;
             } else {
	       switch(oldz) {
                  case PAIR :
                     if(bDel[0][oldrightB-rightB]+logUewk[PAIR][DELETION]+sm->logfreqofstate((*pro_A)[rightA+1])>
	                bPair[0][oldrightB-rightB-1]+logUewk[PAIR][PAIR]+sm->logfreqofstate((*pro_A)[rightA+1])+sm->logProb_ij((*pro_A)[rightA+1],(*pro_B)[rightB+1])) {
                       if(bIns[1][oldrightB-rightB-1]+logUewk[PAIR][INSERTION]+sm->logfreqofstate((*pro_B)[rightB+1])>
		          bDel[0][oldrightB-rightB]+logUewk[PAIR][DELETION]+sm->logfreqofstate((*pro_A)[rightA+1])) {
                          newz=INSERTION;
			  ++rightB;
		       } else {
                          newz=DELETION;
			  ++rightA;
		       }
	             } else {
                       if(bIns[1][oldrightB-rightB-1]+logUewk[PAIR][INSERTION]+sm->logfreqofstate((*pro_B)[rightB+1])>
		          bPair[0][oldrightB-rightB-1]+logUewk[PAIR][PAIR]+sm->logfreqofstate((*pro_A)[rightA+1])+sm->logProb_ij((*pro_A)[rightA+1],(*pro_B)[rightB+1])) {
                          newz=INSERTION;
			  ++rightB;
		       } else {
                          newz=PAIR;
			  ++rightA;
			  ++rightB;
		       }
	             }
	             break;
	          case INSERTION :
                     if(bDel[0][oldrightB-rightB]+logUewk[INSERTION][DELETION]+sm->logfreqofstate((*pro_A)[rightA+1])>
	                bPair[0][oldrightB-rightB-1]+logUewk[INSERTION][PAIR]+sm->logfreqofstate((*pro_A)[rightA+1])+sm->logProb_ij((*pro_A)[rightA+1],(*pro_B)[rightB+1])) {
                       if(bIns[1][oldrightB-rightB-1]+logUewk[INSERTION][INSERTION]+sm->logfreqofstate((*pro_B)[rightB+1])>
		          bDel[0][oldrightB-rightB]+logUewk[INSERTION][DELETION]+sm->logfreqofstate((*pro_A)[rightA+1])) {
                          newz=INSERTION;
			  ++rightB;
		       } else {
                          newz=DELETION;
			  ++rightA;
		       }
	             } else {
                       if(bIns[1][oldrightB-rightB-1]+logUewk[INSERTION][INSERTION]+sm->logfreqofstate((*pro_B)[rightB+1])>
		          bPair[0][oldrightB-rightB-1]+logUewk[INSERTION][PAIR]+sm->logfreqofstate((*pro_A)[rightA+1])+sm->logProb_ij((*pro_A)[rightA+1],(*pro_B)[rightB+1])) {
                          newz=INSERTION;
			  ++rightB;
		       } else {
                          newz=PAIR;
			  ++rightA;
			  ++rightB;
		       }
	             }
	             break;
	          case DELETION :
                     if(bDel[0][oldrightB-rightB]+logUewk[DELETION][DELETION]+sm->logfreqofstate((*pro_A)[rightA+1])>
	                bPair[0][oldrightB-rightB-1]+logUewk[DELETION][PAIR]+sm->logfreqofstate((*pro_A)[rightA+1])+sm->logProb_ij((*pro_A)[rightA+1],(*pro_B)[rightB+1])) {
                       if(bIns[1][oldrightB-rightB-1]+logUewk[DELETION][INSERTION]+sm->logfreqofstate((*pro_B)[rightB+1])>
		          bDel[0][oldrightB-rightB]+logUewk[DELETION][DELETION]+sm->logfreqofstate((*pro_A)[rightA+1])) {
                          newz=INSERTION;
			  ++rightB;
		       } else {
                          newz=DELETION;
			  ++rightA;
		       }
	             } else {
                       if(bIns[1][oldrightB-rightB-1]+logUewk[DELETION][INSERTION]+sm->logfreqofstate((*pro_B)[rightB+1])>
		          bPair[0][oldrightB-rightB-1]+logUewk[DELETION][PAIR]+sm->logfreqofstate((*pro_A)[rightA+1])+sm->logProb_ij((*pro_A)[rightA+1],(*pro_B)[rightB+1])) {
                          newz=INSERTION;
			  ++rightB;
		       } else {
                          newz=PAIR;
			  ++rightA;
			  ++rightB;
		       }
	             }
	             break;
                  default :
                     cerr << "Impossible state in hirschberg II\n";
#ifdef DEBUG
                    cerr<<"Aha: \n";
                    for(int i=0;i<4;++i) {
                       for(int j=0;j<5;++j)
                          cerr<<logUewk[i][j]<<" ";
                       cerr<<endl;  
                    }
#endif
	             exit(EXIT_FAILURE);
               }
           }
	   oldz=newz;//cout<<(int)newz<<" ";
	   rightvec.push_back(newz);
        }//cout<<endl;

       for(unsigned int i=leftvec.size();i>0;--i)
        lokvec.push_back(leftvec[i-1]);
       lokvec.push_back(bz);
       for(unsigned int i=0;i<rightvec.size();++i)
           lokvec.push_back(rightvec[i]);

       rightvec.clear();leftvec.clear();
       bDel.clear();bIns.clear();bPair.clear();
       fDel.clear();fIns.clear();fPair.clear();
       durchDel.clear();durchIns.clear();durchPair.clear();
       // Rekursion
       vector<zustand> hl, hr;
       hirschberg(von_A,von_B,leftA,leftB,hl,firststate,lokvec.front());
       hirschberg(rightA,rightB,bis_A,bis_B,hr,lokvec.back(),laststate);
       for(unsigned int i=0;i<hl.size();++i)
           gesamtvec.push_back(hl[i]);
       for(unsigned int i=0;i<lokvec.size();++i)
           gesamtvec.push_back(lokvec[i]);
       for(unsigned int i=0;i<hr.size();++i)
           gesamtvec.push_back(hr[i]);
    } else {
        if(rightB-leftB>=0 && rightA-leftA>=0) {//cout<<"n: ";
           vector<zustand> zust_vekt;
	   zustand nextstate=laststate;
           align_graph graph(*this,von_A,von_B,bis_A,bis_B,firststate,laststate);//cout<<(int)laststate<<"\t";
           int i=0, j=0;
           int PA=-1, PB=-1;
           switch (laststate) {
           case ENDE :
	       i=pro_A->laen();
	       j=pro_B->laen();
	       break;
           case PAIR :
	       i=bis_A;
	       j=bis_B;
	       PA=(*pro_A)[i];
	       PB=(*pro_B)[j];
	       break;
           case INSERTION :
	       i=bis_A+1;
	       j=bis_B;
	       PB=(*pro_B)[j];
	       break;
           case DELETION :
	       i=bis_A;
	       j=bis_B+1;
	       PA=(*pro_A)[i];
	       break;
           default:
	       cerr << "unmoeglicher \"nextstate\" in hirschberg(...)\n";
           }
           while( i>leftA || j>leftB ) {
 	       nextstate=graph.maxwk_schritt(nextstate,i,j,PA,PB);
	       switch (nextstate) {
	       case PAIR : --i; --j; PA=(*pro_A)[i]; PB=(*pro_B)[j]; break;
	       case INSERTION : --j; PA=-1; PB=(*pro_B)[j]; break;
	       case DELETION : --i; PA=(*pro_A)[i]; PB=-1; break;
	       default: cerr << "unmoeglicher \"nextstate\" in hirschberg!!!\n";
	       }
	       if ( i<von_A || j<von_B )
	           cerr << "Bereich verlassen in  alignment::hirschberg(...)\n";
	       zust_vekt.push_back(nextstate);//cout<<(int)nextstate<<" ";
          }//cout<<endl;

          for(unsigned int i=zust_vekt.size();i>0;--i)
	      gesamtvec.push_back(zust_vekt[i-1]);
	} else { // nur eine der beiden for-Schleifen wird auch ausgefuehrt
            for(int i=0;i<=rightA-leftA;++i)
                gesamtvec.push_back(DELETION);
            for(int i=0;i<=rightB-leftB;++i)
                gesamtvec.push_back(INSERTION);
	   // cout<<"s: ";
	    //for(unsigned int i=0;i<gesamtvec.size();++i)
	        //cout<<gesamtvec[i]<<" ";
	    //cout<<endl;
        }
    }

#ifdef DEBUG
    int aidx=0, bidx=0;
    for(unsigned int i=0;i<gesamtvec.size();++i) {
        switch(gesamtvec[i]) {
            case PAIR :
	      ++aidx; ++bidx;
	      break;
	    case INSERTION :
	      ++bidx;
	      break;
	    case DELETION :
	      ++aidx;
	      break;
	    default :
	      cerr << "Impossible state in Hirschberg algorithm\n";
	      exit(EXIT_FAILURE);
        }
    }
    if(((aidx!=oldrightA-von_A && oldrightA>von_A) || (bidx!=oldrightB-von_B && oldrightB>von_B))) {
         cerr<<"verw: "<<aidx<<" "<<bidx<<endl
             <<von_A<<" "<<oldrightA<<" "<<von_B<<" "<<oldrightB<<endl;
         exit(EXIT_FAILURE);
      }
#endif

}

void alignment::teilsample(int von_A, int von_B, int bis_A, int bis_B) 
{
if(bis_A-von_A>1 && bis_B-von_B>1) { // fuer leere Sequenzen wird nicht neu gesampelt
    /* sample ein neues Teilalignment von (von_A,von_B)
     * bis (bis_A,bis_B). Dabei bleiben die Zustaende der von
     * Start- und Endpunkt erhalten.
     */
    if(log((double)(bis_A-von_A))+log((double)(bis_B-von_B))>log((double)MAXSEQLEN)+log((double)MAXSEQLEN)) {
        hirschberg(FALSE);
    } else {
        zustand nextstate=get_zust(bis_A,bis_B);
        vector<zustand> zust_vekt;
        int a=bis_A, b=bis_B;
        int anz_insert=0,anz_delet=0;
        while (iter_zust(&a,&b)==INSERTION) ++anz_insert;
        if (anz_insert==0) {
	    a=bis_A;
	    b=bis_B;
	    while (iter_zust(&a,&b)==DELETION) ++anz_delet;
        }
        align_graph graph(*this,von_A,von_B,bis_A,bis_B);
        int i=0, j=0;
        int PA=-1, PB=-1;
        switch (nextstate) {
        case ENDE :
	    i=pro_A->laen();
	    j=pro_B->laen();
	    break;
        case PAIR :
	    i=bis_A;
	    j=bis_B;
	    PA=(*pro_A)[i];
	    PB=(*pro_B)[j];
	    break;
        case INSERTION :
	    i=bis_A+1;
	    j=bis_B;
	    PB=(*pro_B)[j];
	    break;
        case DELETION :
	    i=bis_A;
	    j=bis_B+1;
	    PA=(*pro_A)[i];
	    break;
        default:
	    cerr << "unmoeglicher \"nextstate\" in teilsample(...)\n";
        }
        while( i>von_A+1 || j>von_B+1 ) {
            if(Schritt<BISHIERHINZUFAELLIG)
	       nextstate=graph.zufallsschritt(nextstate,i,j,PA,PB,coolexp,Schritt);
	    else
	       nextstate=graph.maxwk_schritt(nextstate,i,j,PA,PB);
	    switch (nextstate) {
	    case PAIR : --i; --j; PA=(*pro_A)[i]; PB=(*pro_B)[j]; break;
	    case INSERTION : --j; PA=-1; PB=(*pro_B)[j]; break;
	    case DELETION : --i; PA=(*pro_A)[i]; PB=-1; break;
	    default: cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	    }
	    if ( i<von_A || j<von_B )
	        cerr << "Bereich verlassen in  alignment::teilsample(...)\n";
	    zust_vekt.push_back(nextstate);
        }
        a=von_A;
        b=von_B;
        for( i=zust_vekt.size()-1; i>=0; --i) {
	    switch(zust_vekt[i]) {
	    case PAIR :
	        ++a;
	        ++b;
	        pAB[a]=b;
	        pBA[b]=a;
	        break;
	    case INSERTION :
	        ++b;  
	        pBA[b]=a;
	        break;
	    case DELETION :
	        ++a;
	        pAB[a]=b;
	        break;
	    case ENDE :
	        break;
	    default:
	        cerr << "unmoeglicher \"nextstate\" in teilsample!!!\n";
	    }
        }
        for( i=1; i<anz_insert; ++i){
	    pBA[bis_B+i]=pBA[bis_B];
        }
        for( i=1; i<anz_delet; ++i){
	    pAB[bis_A+i]=pAB[bis_A];
        }
    }
  }
}


void alignment::teilsample(int von_A, int bis_A) {
    /* sample ein neues Teilalignment von dna_A[von_A]
     * bis dna_A[bis_A].
   */
    if(pAB[von_A]==pAB[bis_A])
       sample();
    else
       teilsample(von_A,pAB[von_A],bis_A,pAB[bis_A]);
}

void alignment::teilsample(int laenge) {
    /* sample ein neues Teilalignment zu einem zufaelligen Abschnitt der
     * Laenge laenge. 
   */
    if (laenge<1) cerr << "unsinniger Parameter in alignment::teilsample(...)\n";
    if (laenge>=pro_A->laen() ) sample();
    else {
	int R=rand()%(pro_A->laen()+laenge);
	if ( R >= pro_A->laen() ){
            if(pAB[laenge]==-1)   // hat die zweite Sequenz noch nicht begonnen, wird der ganze Pfad gesampelt
	       sample();
	    else
	       teilsample( -1,-1, laenge, pAB[laenge]);
	} else {
	    if ( R+laenge >= pro_A->laen() ) {
	        if(pAB[pro_A->laen()-laenge]==pro_B->laen()-1)// ist die zweite Sequenz schon zuende, wird der ganze Pfad gesampelt
		   sample();
		else
		   teilsample( pro_A->laen()-laenge, pAB[pro_A->laen()-laenge],
			       pro_A->laen(),pro_B->laen());
	    } else {
		teilsample( R,R+laenge);
	    }
	}
    }
}


zustand alignment::get_zust(int a,int b) const {
    /* Liefert den zu a und b assoziierten Alignment-Zustand
     */
    if (a==-1) {
	if (b==-1) 
	    return START;
	if (pBA[b]==-1) 
	    return INSERTION;
	return NOSTATE;
    }
    if (a==pro_A->laen()) {
	if (b==pro_B->laen())
	    return ENDE;
	return NOSTATE;
    }
    if (b==-1) {
	if (pAB[a]==-1) 
	    return DELETION;
	return NOSTATE;
    }
    if (b==pro_B->laen()) {
	return NOSTATE;
    }
    if (b==pAB[a]){
	if (a==pBA[b])
	    return PAIR;
	if (pBA[b]<a)
	    return DELETION;
	return NOSTATE;
    }
    if (pBA[b]==a && pAB[a]<b )
	return INSERTION;
    return NOSTATE;
}

zustand alignment::iter_zust(int* pa,int* pb) const {
    /* Liefert den zu *pa und *pb assoziierten Alignment-Zustand und 
     * setzt (*pa,*pb) einen Zustand weiter.
   */
    zustand zust;
    zust=get_zust(*pa,*pb);
    switch (zust) {
    case START:
    case PAIR:  
    case INSERTION: 
    case DELETION:
	if ( *pa==pro_A->laen()-1 ) {
	    ++*pb;
	    if ( *pb==pro_B->laen() ) 
		++*pa;
	}
	else {
	    if ( *pb==pro_B->laen()-1 )
		++*pa;
	    else {
		if ( pAB[*pa+1]==*pb )
		    ++*pa;
		else {
		    ++*pb;
		    if ( pBA[*pb]!=*pa ) 
			++*pa;
		}
	    }
	}
	break;
    case ENDE:
	break;
    default:
	cerr << *pa << " und " << *pb << " ergeben keinen Zustand"
	     << " in alignment::iter_zust\n";
	exit(1);
    }
    return zust;
}

void alignment::zurueck(baum *b_1, baum *b_2)
// schreibt das ermittelte Alignment zurueck in den Baum
{
   // sind b_1 und b_2 Nachbarn
   int idxA = b_1->whichone(b_2);
   int idxB = b_2->whichone(b_1);
   if(pBA.size()!=0 && pAB.size()!=0) {
      // stimmt auch alles?
      if(pBA.size()!=b_2->pro.size() || pAB.size()!=b_1->pro.size() ||
        ((pBA.back()!=b_1->pro.laen()-1 || pAB.back()>b_2->pro.laen()-1) && 
         (pAB.back()!=b_2->pro.laen()-1 || pBA.back()>b_1->pro.laen()-1)))
      {
          cerr << "Found inconsistencies while writing an alignment!\n"
               << pBA.size() << " " << b_2->pro.size()<<" "<<pAB.back()<<endl
               << pAB.size() << " " << b_1->pro.size()<<" "<<pBA.back()<<endl;
          exit(EXIT_FAILURE);
      }
   } else {
      if(pBA.size()!=pro_B->size() || pAB.size()!=pro_A->size())
      {
          cerr << "Found inconsistencies while constructing an alignment!\n"
               << pBA.size() << " " << pro_B->size() <<endl 
               << pAB.size() << " " << pro_A->size() <<endl; 
          exit(EXIT_FAILURE);
      } else 
          cerr << "Warning:there are internal sequences with length 0\n";
   }
   
   b_1->nb[idxA].ausrichter=pAB;
   b_2->nb[idxB].ausrichter=pBA;
}


double alignment::EMloglikl(double time,double scale)
{
  unsigned int i, j;
  double retval=0.0; 
  sm->settime(time);
  idm->fillP(time*scale);

  for(i=0;i<hmmtrans.size();++i)
      for(j=0;j<hmmtrans[i].size();++j)
          if(hmmtrans[i][j]>0)
	     retval+=hmmtrans[i][j]*(idm->retlnprob(i,j));
  for(i=0;i<basep.size();++i)
      for(j=0;j<basep[i].size();++j)
	  retval+=basep[i][j]*(sm->logProb_ij(i,j));

  return -retval; // wg. Minimierung
}



#define ITMAX 100
#define CGOLD 0.3819660
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define ZEPS 1.0e-10
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

/* Brents method in one dimension */
double alignment::brent(double ax, double bx, double cx, double tol,
	double *foptx, double *f2optx, double fax, double fbx, double fcx, double otherarg, bool whicharg)
{
	int iter;
	double a,b,d=0,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double xw,wv,vx;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=bx;
	fx=fbx;
	if (fax < fcx) {
		w=ax;
		fw=fax;
		v=cx;
		fv=fcx;
	} else {
		w=cx;
		fw=fcx;
		v=ax;
		fv=fax;	
	}
	for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*foptx = fx;
			xw = x-w;
			wv = w-v;
			vx = v-x;
			*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
				(v*v*xw + x*x*wv + w*w*vx);
			return x;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		if(!whicharg)
		   fu=EMloglikl(u,otherarg);
		else
		   fu=EMloglikl(otherarg,u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	*foptx = fx;
	xw = x-w;
	wv = w-v;
	vx = v-x;
	*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
		(v*v*xw + x*x*wv + w*w*vx);
	return x;
}
#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SHFT
#undef SIGN
#undef GOLD
#undef GLIMIT
#undef TINY


/* one-dimensional minimization - as input a lower and an upper limit and a trial
  value for the minimum is needed: xmin < xguess < xmax
  the function and a fractional tolerance has to be specified
  onedimenmin returns the optimal x value and the value of the function
  and its second derivative at this point
  */
void alignment::onedimenmin(double xmin, double xguess, double xmax, double& optx,
	double tol, double *fx, double *f2x, double otherarg, bool whicharg)
{
	double eps, ax, bx, cx, fa, fb, fc;
		
	/* first attempt to bracketize minimum */
	eps = xguess*tol*50.0;
	ax = xguess - eps;
	if (ax < xmin) ax = xmin;
	bx = xguess;
	cx = xguess + eps;
	if (cx > xmax) cx = xmax;
	
	/* check if this works */
	if(!whicharg) {
	   fa = EMloglikl(ax,otherarg);
	   fb = EMloglikl(bx,otherarg);
	   fc = EMloglikl(cx,otherarg);
	} else {
	   fa = EMloglikl(otherarg,ax);
	   fb = EMloglikl(otherarg,bx);
	   fc = EMloglikl(otherarg,cx);
	}

	/* if it works use these borders else be conservative */
	if ((fa < fb) || (fc < fb)) {
		if (ax != xmin) {
		    if(!whicharg)
		       fa = EMloglikl(xmin,otherarg);
		    else
		       fa = EMloglikl(otherarg,xmin);
		}
		if (cx != xmax) {
		    if(!whicharg)
		       fc = EMloglikl(xmax,otherarg);
		    else
		       fc = EMloglikl(otherarg,xmax);
		}
		optx = brent(xmin, xguess, xmax, tol, fx, f2x, fa, fb, fc,otherarg,whicharg);
	} else
		optx = brent(ax, bx, cx, tol, fx, f2x, fa, fb, fc,otherarg,whicharg);
}


/* two-dimensional minimization with borders and calculations of standard errors */
/* we optimize along basis vectors - not very optimal but it seems to work well */
void alignment::twodimenmin(double tol,
	int active1, double min1, double& x1, double max1, double *err1,
	int active2, double min2, double& x2, double max2, double *err2)
{
	int it, nump, change;
	double x1old, x2old;
	double fx, f2x;

	it = 0;
	nump = 0;

	/* count number of parameters */
	if (active1) nump++;
	if (active2) nump++;

	do { /* repeat until nothing changes any more */
		it++;
		change = FALSE;

		/* optimize first variable */
		if (active1) {

			if (x1 <= min1) x1 = min1 + 0.2*(max1-min1);
			if (x1 >= max1) x1 = max1 - 0.2*(max1-min1);
			x1old = x1;
			onedimenmin(min1, x1old, max1, x1, tol, &fx, &f2x, x2, FALSE);
			if (x1 < min1) x1 = min1;
			if (x1 > max1) x1 = max1;
			/* same tolerance as 1D minimization */
			if (fabs(x1 - x1old) > 3.3*tol) change = TRUE;
			
			/* standard error */
			f2x = fabs(f2x);
			if (1.0/(max1*max1) < f2x) (*err1) = sqrt(1.0/f2x);
			else (*err1) = max1;

		}

		/* optimize second variable */
		if (active2) {

			if (x2 <= min2) x2 = min2 + 0.2*(max2-min2);
			if (x2 >= max2) x2 = max2 - 0.2*(max2-min2);
			x2old = x2;
			onedimenmin(min2, x2old, max2, x2, tol, &fx, &f2x, x1, TRUE);
			if (x2 < min2) x2 = min2;
			if (x2 > max2) x2 = max2;
			/* same tolerance as 1D minimization */
			if (fabs(x2 - x2old) > 3.3*tol) change = TRUE;
			
			/* standard error */
			f2x = fabs(f2x);
			if (1.0/(max2*max2) < f2x) (*err2) = sqrt(1.0/f2x);
			else (*err2) = max2;

		}

		if (nump == 1) return;
		
	} while (it != MAXITS && change);

	return;
}


void alignment::optimal_time(double& time, double& scale)
// optimiert die Zeit fuer 2 Sequenzen und gegebenes Modell
{
   int i=0;
   double opttime=time, optscale=scale, err1, err2, oldtime=0.0, oldscale=0.0;
   
   hmmtrans.resize(4);
   for(unsigned int j=0;j<hmmtrans.size();++j)
       hmmtrans[j].resize(5);
   alphsize alf=sm->retalf();
   basep.resize(alf);
   for(unsigned int j=0;j<basep.size();++j)
       basep[j].resize(alf);
//cout<<"Anfang: "<<time<<" "<<scale<<endl;
   bool flag=TRUE;
   do {
     if(i!=0) {
        oldtime=time;
	oldscale=scale;
        time=opttime;
	scale=optscale;
     }
     sample_paths();
     //onedimenmin(MIN_TIME,time,MAX_TIME,opttime,EPSILON,&fx,&fx2);
     twodimenmin(EPSILON,
                 1,MIN_TIME,opttime,MAX_TIME,&err1,
		 1,MIN_SCALE,optscale,MAX_SCALE,&err2);//cout<<time-opttime<<" "<<scale-optscale<<endl;
     ++i;
     if(((time>opttime-EMPREC && time<opttime+EMPREC) && (scale>optscale-EMPREC && scale<optscale+EMPREC)) ||
        ((oldtime>opttime-EMPREC && oldtime<opttime+EMPREC) && (oldscale>optscale-EMPREC && oldscale<optscale+EMPREC))) // sonst kommt's bei Viterbi-Training zu 'flip-flops'
	  flag=FALSE;
   }while(i<EMSTOP && flag);//cout<<"Ende:   "<<time<<" "<<scale<<endl;
   //if(i==EMSTOP)
   //   cerr<<"Warning: One pairwise distance did not converge within "<<EMSTOP<<" EM steps.\n";
   time=opttime;
   scale=optscale;

}

void alignment::optimal_time(double& time)
// optimiert die Zeit fuer 2 Sequenzen und gegebenes Modell
{
   int i=0;
   double ooldtime=DBL_MIN, oldtime=DBL_MIN, fx, fx2;

   hmmtrans.resize(4);
   for(unsigned int j=0;j<hmmtrans.size();++j)
       hmmtrans[j].resize(5);
   alphsize alf=sm->retalf();
   basep.resize(alf);
   for(unsigned int j=0;j<basep.size();++j)
       basep[j].resize(alf);
   bool flag=TRUE;
   do {
     if(i>0)
       ooldtime=oldtime;
     oldtime=time;
     sample_paths();
     onedimenmin(MIN_TIME,oldtime,MAX_TIME,time,EPSILON,&fx,&fx2,1.0,FALSE);
     ++i;
     if((time>oldtime-EMPREC && time<oldtime+EMPREC) ||
        (time>ooldtime-EMPREC && time<ooldtime+EMPREC)) // verhindert 'flip-flops'
       flag=FALSE;
   }while(i<EMSTOP && flag);
   //if(i==EMSTOP) {
   //   cerr<<"Warning: One pairwise distance did not converge within "<<EMSTOP<<" EM steps.\n";
   //}
}

double alignment::opt_for_giv()
{
   double retval=sm->timeOf();
   double oldtime, fx, fx2;
   
   hmmtrans.resize(4);
   for(unsigned int j=0;j<hmmtrans.size();++j)
       hmmtrans[j].resize(5);
   alphsize alf=sm->retalf();
   basep.resize(alf);
   for(unsigned int j=0;j<basep.size();++j)
       basep[j].resize(alf);
   for(unsigned int j=0;j<hmmtrans.size();++j)
     for(unsigned int k=0;k<hmmtrans[j].size();++k)
	  hmmtrans[j][k]=0.0;
   for(int j=0;j<alf;++j)
    for(int k=0;k<alf;++k)
	basep[j][k]=0.0;
   int i=-1; int j=-1;
   int oldi=i, oldj=j;
   int PA=-1, PB=-1;
   zustand zus=iter_zust(&i,&j), zusalt=START;
   while(zus!=ENDE) {
      zusalt=zus;
      zus=iter_zust(&i,&j);
      hmmtrans[zusalt][zus]+=1.0;
      switch (zus) {
	    case PAIR :
	         PA=(*pro_A)[oldi]; PB=(*pro_B)[oldj];
                 // aus Faulheit wird im folgenden nicht
		 // zwischen den verschiedenen Wildcards unterschieden!!!!
	         if(PA<alf) {
	            if(PB<alf) {
		      basep[PA][PB]+=1.0;
		    } else {
		      for(int k=0;k<alf;++k)
		        basep[PA][k]+=1.0/(double)alf;
		    }
		 } else {
	            if(PB<alf) {
		      for(int k=0;k<alf;++k)
		          basep[k][PB]+=1.0/(double)alf;
		    } else {
		      for(int k=0;k<alf;++k)
		        for(int l=0;l<alf;++l)
		          basep[k][l]+=1.0/(double)(alf*alf);
		    }
	         }
		 break;
	    case INSERTION : PA=-1; PB=(*pro_B)[oldj]; break;
	    case DELETION : PA=(*pro_A)[oldi]; PB=-1; break;
	    case ENDE : break;
	    default: cerr << "unmoeglicher \"nextstate\" "<<(int)zus<<" in opt_for_giv!!!\n";
	             exit(EXIT_FAILURE);
       }
       oldi=i;oldj=j;
   }
   //do {
   oldtime=retval;
   onedimenmin(MIN_TIME,oldtime,MAX_TIME,retval,EPSILON,&fx,&fx2,1.0,FALSE);
   //  ++z;
   //}while(i<EMSTOP && (retval<oldtime-EMPREC || retval>oldtime+EMPREC));
   //if(z==EMSTOP) {
   //   cerr<<"Warning: One pairwise distance did not converge within "<<EMSTOP<<" EM steps.\n";
   //}
   return retval;
}

void alignment::sample_paths() 
{
   if(log((double)(pro_A->laen()+1))+log((double)(pro_B->laen()+1))<=log((double)MAXSEQLEN)+log((double)MAXSEQLEN)) {
      alphsize alf=sm->retalf();

      for(unsigned int j=0;j<hmmtrans.size();++j)
          for(unsigned int k=0;k<hmmtrans[j].size();++k)
            // if(j==START || k==ENDE)
	        hmmtrans[j][k]=0.0;
	     //else
	        //hmmtrans[j][k]=1.0; // einfacher Pseudocount
      for(int j=0;j<alf;++j)
         for(int k=0;k<alf;++k)
	     basep[j][k]=0.0;
	     //basep[j][k]=1.0;   // einfacher Pseudocount

      align_graph graph(*this,-1,-1,pro_A->laen(),pro_B->laen());
      int count=0;
      zustand zusalt, zus;
      while(count<EMSAMP) {
          zusalt=ENDE; zus=ENDE;
          int i=pro_A->laen(),j=pro_B->laen();
          int PA=-1, PB=-1;
          while(i>0 || j>0){
            zus=graph.zufallsschritt(zusalt,i,j,PA,PB,1,0);
            hmmtrans[zus][zusalt]+=1.0;
	    switch (zus) {
	       case PAIR :
	            --i; --j; PA=(*pro_A)[i]; PB=(*pro_B)[j];
                    // aus Faulheit wird im folgenden nicht
		    // zwischen den verschiedenen Wildcards unterschieden!!!!
	            if(PA<alf) {
	               if(PB<alf) {
		         basep[PA][PB]+=1.0;
		       } else {
		         for(int k=0;k<alf;++k)
		           basep[PA][k]+=1.0/(double)alf;
		       }
		    } else {
	               if(PB<alf) {
		         for(int k=0;k<alf;++k)
		             basep[k][PB]+=1.0/(double)alf;
		       } else {
		         for(int k=0;k<alf;++k)
		           for(int l=0;l<alf;++l)
		             basep[k][l]+=1.0/(double)(alf*alf);
		       }
	            }
		    break;
	       case INSERTION : --j; PA=-1; PB=(*pro_B)[j]; break;
	       case DELETION : --i; PA=(*pro_A)[i]; PB=-1; break;
	       default: cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	    }
       
            zusalt=zus;
          }
          hmmtrans[START][zus]+=1.0;
          ++count;
      }
      for(unsigned int i=0;i<basep.size();++i)
        for(unsigned int j=0;j<basep[i].size();++j)
          basep[i][j]/=(double)EMSAMP;
      for(unsigned int i=0;i<hmmtrans.size();++i)
        for(unsigned int j=0;j<hmmtrans[i].size();++j)
          hmmtrans[i][j]/=(double)EMSAMP;
   } else {  // Speichersparen
       hirschberg(TRUE);
   }
}


